from collections import defaultdict
from typing import Any, Dict, List, Sequence, Tuple

from pflows.typedef import Annotation, Image


def calculate_iou(
    bbox1: Tuple[float, float, float, float], bbox2: Tuple[float, float, float, float]
) -> float:
    # Calculate the coordinates of the intersection rectangle
    x1 = max(bbox1[0], bbox2[0])
    y1 = max(bbox1[1], bbox2[1])
    x2 = min(bbox1[2], bbox2[2])
    y2 = min(bbox1[3], bbox2[3])

    # Calculate the area of the intersection rectangle
    intersection_area = max(0, x2 - x1) * max(0, y2 - y1)

    # Calculate the area of both bounding boxes
    bbox1_area = (bbox1[2] - bbox1[0]) * (bbox1[3] - bbox1[1])
    bbox2_area = (bbox2[2] - bbox2[0]) * (bbox2[3] - bbox2[1])

    # Calculate the IoU
    iou = intersection_area / (bbox1_area + bbox2_area - intersection_area)

    return iou


def find_best_match(
    gold_annotation: Annotation, new_annotations: List[Annotation]
) -> Tuple[Annotation | None, float]:
    best_match = None
    best_iou = 0.0

    for new_annotation in new_annotations:
        if (
            gold_annotation.category_name == new_annotation.category_name
            and gold_annotation.bbox is not None
            and new_annotation.bbox is not None
        ):
            iou = calculate_iou(gold_annotation.bbox, new_annotation.bbox)
            if iou > best_iou:
                best_match = new_annotation
                best_iou = iou

    return (best_match, best_iou)


def compare_annotations(
    gold_annotation: Annotation, infer_annotations: List[Annotation], iou_threshold: float = 0.5
) -> Dict[str, Any]:
    category_name = gold_annotation.category_name

    # Find the best matching annotation in the new image
    best_match, best_iou = find_best_match(gold_annotation, infer_annotations)

    status = "not found"
    confidence = 0.0
    if best_match is not None:
        if best_iou >= iou_threshold:
            status = "found"
            confidence = best_match.conf

    return {
        "id": gold_annotation.id,
        "matched_id": best_match.id if best_match else "",
        "status": status,
        "confidence": confidence,
        "iou": best_iou,
    }


def calculate_metrics(
    original_annotations: List[Annotation],
    inferred_annotations: List[Annotation],
    category_names: Sequence[str],
    iou_threshold: float = 0.5,
) -> Dict[str, Dict[str, float]]:
    metrics = defaultdict(lambda: {"TP": 0, "FP": 0, "FN": 0})
    matched_inferred = set()
    incorrect_annotations = {"FP": [], "FN": []}

    for gold_annotation in original_annotations:
        result = compare_annotations(gold_annotation, inferred_annotations, iou_threshold)
        category = gold_annotation.category_name
        if category not in category_names:
            continue
        if result["status"] == "found":
            metrics[category]["TP"] += 1
            matched_inferred.add(result["matched_id"])
        else:
            metrics[category]["FN"] += 1
            incorrect_annotations["FN"].append(gold_annotation)

    for inferred_annotation in inferred_annotations:
        if inferred_annotation.category_name not in category_names:
            continue
        category = inferred_annotation.category_name
        if inferred_annotation.id not in matched_inferred:
            metrics[category]["FP"] += 1
            incorrect_annotations["FP"].append(inferred_annotation)

    return dict(metrics), incorrect_annotations


def compare_images_annotations(
    gold_images: Sequence[Image],
    inferred_images: Sequence[Image],
    category_names: Sequence[str],
    iou_threshold: float = 0.5,
) -> Dict[str, Any]:
    gold_ids = set(image.id for image in gold_images)
    infer_ids = set(image.id for image in inferred_images)
    infer_ids = infer_ids.union(
        set(image_id for image in inferred_images for image_id in image.intermediate_ids)
    )
    common_ids = gold_ids.intersection(infer_ids)

    overall_metrics = defaultdict(lambda: {"TP": 0, "FP": 0, "FN": 0})
    confusion_matrix = defaultdict(lambda: defaultdict(int))

    incorrect_images = {}

    for image_id in common_ids:
        original_image = next(img for img in gold_images if img.id == image_id)
        inferred_image = next(img for img in inferred_images if img.id == image_id)
        image_metrics, incorrect_annotations = calculate_metrics(
            original_image.annotations, inferred_image.annotations, category_names, iou_threshold
        )

        if incorrect_annotations["FP"] or incorrect_annotations["FN"]:
            incorrect_images[image_id] = incorrect_annotations

        for category, metrics in image_metrics.items():
            overall_metrics[category]["TP"] += metrics["TP"]
            overall_metrics[category]["FP"] += metrics["FP"]
            overall_metrics[category]["FN"] += metrics["FN"]

        # Build confusion matrix
        for gold_ann in original_image.annotations:
            best_match = None
            best_iou = 0
            if gold_ann.category_name not in category_names:
                continue
            for infer_ann in inferred_image.annotations:
                if infer_ann.category_name not in category_names:
                    continue
                if not gold_ann.bbox or not infer_ann.bbox:
                    continue
                iou = calculate_iou(gold_ann.bbox, infer_ann.bbox)
                if iou > best_iou:
                    best_match = infer_ann
                    best_iou = iou

            if best_match and best_iou >= iou_threshold:
                confusion_matrix[gold_ann.category_name][best_match.category_name] += 1
            else:
                confusion_matrix[gold_ann.category_name]["not_detected"] += 1

    category_metrics = {}
    for category, metrics in overall_metrics.items():
        tp, fp, fn = metrics["TP"], metrics["FP"], metrics["FN"]
        precision = tp / (tp + fp) if (tp + fp) > 0 else 0
        recall = tp / (tp + fn) if (tp + fn) > 0 else 0
        f1_score = (
            2 * (precision * recall) / (precision + recall) if (precision + recall) > 0 else 0
        )

        category_metrics[category] = {
            "precision": precision,
            "recall": recall,
            "f1_score": f1_score,
            "TP": tp,
            "FP": fp,
            "FN": fn,
        }

    total_tp = sum(metrics["TP"] for metrics in overall_metrics.values())
    total_fp = sum(metrics["FP"] for metrics in overall_metrics.values())
    total_fn = sum(metrics["FN"] for metrics in overall_metrics.values())

    overall_precision = total_tp / (total_tp + total_fp) if (total_tp + total_fp) > 0 else 0
    overall_recall = total_tp / (total_tp + total_fn) if (total_tp + total_fn) > 0 else 0
    overall_f1_score = (
        2 * (overall_precision * overall_recall) / (overall_precision + overall_recall)
        if (overall_precision + overall_recall) > 0
        else 0
    )

    return {
        "raw": overall_metrics,
        "categories": category_metrics,
        "overall": {
            "precision": overall_precision,
            "recall": overall_recall,
            "f1_score": overall_f1_score,
            "TP": total_tp,
            "FP": total_fp,
            "FN": total_fn,
        },
        "confusion_matrix": confusion_matrix,
        "incorrect_images": incorrect_images,
    }



from typing import Any, Dict, List, Optional, Sequence, Tuple
import os
from uuid import uuid4
from concurrent.futures import ThreadPoolExecutor, as_completed

import cv2
import numpy as np

import albumentations as A
from shapely.geometry import Polygon

from pflows.polygons import bbox_from_polygon, calculate_center_from_polygon
from pflows.typedef import Annotation, Dataset, Image


def find_biggest_contour(
    contours: Sequence[np.ndarray[np.float32, np.dtype[np.float32]]]
) -> Optional[np.ndarray[np.float32, np.dtype[np.float32]]]:
    max_area: float = 0
    max_contour = None
    for contour in contours:
        area = cv2.contourArea(contour)
        if area > max_area:
            max_area = area
            max_contour = contour
    return max_contour


def get_new_polygon(
    transformed_mask: Any, polygon_id: int, width: int, height: int
) -> Tuple[float, ...]:
    contours = cv2.findContours(
        (transformed_mask == polygon_id).astype(np.uint8),
        cv2.RETR_EXTERNAL,
        cv2.CHAIN_APPROX_SIMPLE,
    )[0]
    contour = find_biggest_contour(contours)
    if contour is None:
        raise ValueError("No contour found")
    polygon = []
    epsilon = 0.003 * cv2.arcLength(contour, True)
    smoothed_contour = cv2.approxPolyDP(contour, epsilon, True)
    try:
        for point in smoothed_contour.squeeze():
            polygon.append(point[0])
            polygon.append(point[1])
    # pylint: disable=broad-exception-caught
    except Exception:
        for point in contour.squeeze():
            polygon.append(point[0])
            polygon.append(point[1])

    return tuple(
        round(p / width, 5) if i % 2 == 0 else round(p / height, 5) for i, p in enumerate(polygon)
    )


def get_new_polygons(transformed_mask: Any, width: int, height: int) -> List[Tuple[float, ...]]:
    new_polygons = []
    for polygon_id in np.unique(transformed_mask):
        if polygon_id > 0:
            new_polygon = get_new_polygon(transformed_mask, polygon_id, width, height)
            new_polygons.append(new_polygon)
    return new_polygons


def apply_augmentation_with_multiple_masks(image, polygons, transform):
    height, width = image.shape[:2]

    # Create a single multi-channel mask
    num_polygons = len(polygons)
    multi_mask = np.zeros((height, width, num_polygons), dtype=np.uint8)

    for index, polygon_data in enumerate(polygons):
        mask = np.zeros((height, width), dtype=np.uint8)
        pts = np.array(
            [polygon_data[i : i + 2] for i in range(0, len(polygon_data), 2)], dtype=np.int32
        )
        cv2.fillPoly(mask, [pts], 1)
        multi_mask[:, :, index] = mask

    # Apply the transformation
    transformed = transform(image=image, masks=[multi_mask])
    transformed_image = transformed["image"]
    transformed_multi_mask = transformed["masks"][0]

    # Get new polygons from each channel of the transformed multi-mask
    new_polygons = []
    for channel in range(num_polygons):
        new_polygons.extend(get_new_polygons(transformed_multi_mask[:, :, channel], width, height))

    return transformed_image, new_polygons


# pylint: disable=too-many-locals,too-many-arguments
def generate_augmented_image(
    image_id: str,
    image_path: str,
    segments: List[Tuple[float, ...]],
    target_folder: str,
    new_index: int,
    total_index: int,
) -> Optional[Dict[str, Any]]:
    new_id = f"{image_id}_aug_{new_index}"
    target_file = f"{target_folder}/{new_id}.jpg"
    image = cv2.imread(image_path)
    height, width = image.shape[:2]

    retry_number = 5
    new_polygons = []
    transformed_image = None
    for index in range(retry_number):
        if index != 0:
            print(f"\tRetrying {index} for {new_id}")
        polygons: List[List[int]] = []
        for polygon_raw in segments:
            polygon = []
            for i in range(0, len(polygon_raw), 2):
                polygon.extend(
                    [int(width * float(polygon_raw[i])), int(height * float(polygon_raw[i + 1]))]
                )
            polygons.append(polygon)

        new_polygons = []
        transformed_image = None
        try:
            transform = A.Compose(
                [
                    A.Affine(
                        rotate=(-1.5, 1.5),
                        translate_percent=(-0.015, 0.015),
                        scale=(1, 1.2),
                        shear=(-2, 2),
                        p=1.0,
                    ),
                    A.Defocus(radius=(1, 4), alias_blur=(0.1, 0.2), p=0.6),
                    A.CLAHE(clip_limit=2, p=0.9),
                    A.RandomBrightnessContrast(
                        p=0.5, brightness_limit=(-0.1, 0.1), contrast_limit=0.2
                    ),
                    A.MultiplicativeNoise(multiplier=(0.8, 1.2), per_channel=True, p=0.2),
                    A.GaussNoise(var_limit=(20, 80), mean=50, p=0.8),
                    A.ElasticTransform(alpha=1, sigma=25, alpha_affine=25, p=1.0),
                ]
            )

            transformed_image, new_polygons = apply_augmentation_with_multiple_masks(
                image, polygons, transform
            )
        except Exception:
            continue

        if len(new_polygons) == 0:
            print(f"\tFailed {new_id} no polygons")
            continue
        break

    if transformed_image is None:
        print(f"\tFailed {new_id} no transformation")
        return None
    cv2.imwrite(target_file, transformed_image)
    print("image:", total_index)
    return {
        "path": target_file,
        "segments": new_polygons,
        "id": new_id,
        "original_id": image_id,
        "image_index": new_index,
        "total_index": total_index,
    }


def generate_augmentations(images: List[Image], number: int = 3) -> List[Image]:
    random_temp_folder = f"/tmp/tmp_images_augmented/{uuid4()}"
    os.makedirs(random_temp_folder, exist_ok=True)
    new_images = []
    original_images_by_id = {image.id: image for image in images}

    with ThreadPoolExecutor() as executor:
        futures = []

        for index, image in enumerate(images):
            for i in range(number):
                annotation_segments = [
                    annotation.segmentation
                    for annotation in image.annotations
                    if annotation.segmentation is not None
                ]
                futures.append(
                    executor.submit(
                        generate_augmented_image,
                        image.id,
                        image.path,
                        annotation_segments,
                        random_temp_folder,
                        i,
                        index,
                    )
                )

        for future in as_completed(futures):
            # try:
            new_image_info = future.result()
            if not new_image_info:
                continue
            original_image = original_images_by_id[new_image_info["original_id"]]
            new_images.append(
                Image(
                    id=new_image_info["id"],
                    path=new_image_info["path"],
                    intermediate_ids=original_image.intermediate_ids + [original_image.id],
                    width=original_image.width,
                    height=original_image.height,
                    size_kb=original_image.size_kb,
                    group=original_image.group,
                    annotations=[
                        Annotation(
                            id=f"{annotation.id}_{new_image_info['image_index']}",
                            segmentation=new_segmentation,
                            category_id=annotation.category_id,
                            bbox=bbox_from_polygon(new_segmentation),
                            center=calculate_center_from_polygon(new_segmentation),
                            task=annotation.task,
                            conf=annotation.conf,
                            category_name=annotation.category_name,
                            tags=annotation.tags + ["augmented"],
                        )
                        for annotation, new_segmentation in zip(
                            original_image.annotations, new_image_info["segments"]
                        )
                    ],
                ),
            )
            print("Augmented image", new_images[-1].id)

    return new_images


def generic(dataset: Dataset, number: int = 2) -> Dataset:
    training_images = [image for image in dataset.images if image.group == "train"]
    augmented_images = generate_augmentations(training_images, number)
    return Dataset(
        images=dataset.images + augmented_images,
        categories=dataset.categories,
        groups=dataset.groups,
    )


def by_categories(dataset: Dataset, categories_numbers: Dict[str, int]) -> Dataset:
    training_images = [image for image in dataset.images if image.group == "train"]
    new_augmented_images = []
    for category, number in categories_numbers.items():
        category_training = [
            image
            for image in training_images
            if any(annotation.category_name == category for annotation in image.annotations)
        ]
        augmented_images = generate_augmentations(category_training, number)
        new_augmented_images += augmented_images
    return Dataset(
        images=dataset.images + new_augmented_images,
        categories=dataset.categories,
        groups=dataset.groups,
    )



import os
from typing import Tuple
from roboflow import Roboflow


def parse_url(url: str) -> Tuple[str, str, str]:
    path_url = ""
    if "app.roboflow.com" in url:
        path_url = url.split("app.roboflow.com/")[1]
    else:
        path_url = url.split("universe.roboflow.com/")[1]
    user = path_url.split("/")[0]
    project = path_url.split("/")[1]
    dataset_version = path_url.split("/")[-1]
    return user, project, dataset_version


def download_dataset(url: str, target_dir: str) -> bool:
    if os.path.exists(target_dir):
        print()
        print("Dataset already downloaded")
        return False

    if not os.environ.get("ROBOFLOW_API_KEY"):
        raise ValueError("ROBOFLOW_API_KEY environment variable not set, add it to .env file.")
    rf = Roboflow(api_key=os.environ["ROBOFLOW_API_KEY"])
    user, project_name, dataset_version = parse_url(url)

    project = rf.workspace(user).project(project_name)
    project.version(int(dataset_version)).download("yolov8", target_dir)
    return True



from typing import Tuple, Sequence

import numpy as np
from shapely.geometry import Point, Polygon, MultiPolygon
from shapely.validation import make_valid
from shapely.errors import TopologicalError

from pflows.typedef import Annotation

ROUNDING = 6


def calculate_center_from_bbox(bbox: Tuple[float, float, float, float]) -> Tuple[float, float]:
    x1, y1, x2, y2 = bbox
    return (round((x1 + x2) / 2, ROUNDING), round((y1 + y2) / 2, ROUNDING))


def calculate_center_from_polygon(polygon: Tuple[float, ...]) -> Tuple[float, float] | None:
    if len(polygon) == 0:
        return None
    x = [polygon[i] for i in range(0, len(polygon), 2)]
    y = [polygon[i] for i in range(1, len(polygon), 2)]
    return (round(sum(x) / len(x), ROUNDING), round(sum(y) / len(y), ROUNDING))


def polygon_from_bbox(bbox: Tuple[float, float, float, float]) -> Tuple[float, ...]:
    x1, y1, x2, y2 = bbox
    return (x1, y1, x2, y1, x2, y2, x1, y2)


def bbox_from_polygon(polygon: Tuple[float, ...]) -> Tuple[float, float, float, float]:
    x = [polygon[i] for i in range(0, len(polygon), 2)]
    y = [polygon[i] for i in range(1, len(polygon), 2)]
    return (min(x), min(y), max(x), max(y))


def get_bbox(x1: float, y1: float, x2: float, y2: float) -> Tuple[float, float, float, float]:
    return (x1, y1, x2, y2)


def get_min_max_bbox(annotations: list[Annotation]) -> tuple[float, float, float, float] | None:
    if all(annotation.bbox is None for annotation in annotations):
        return None
    x1 = min(annotation.bbox[0] for annotation in annotations if annotation.bbox is not None)
    y1 = min(annotation.bbox[1] for annotation in annotations if annotation.bbox is not None)
    x2 = max(annotation.bbox[2] for annotation in annotations if annotation.bbox is not None)
    y2 = max(annotation.bbox[3] for annotation in annotations if annotation.bbox is not None)
    return get_bbox(x1, y1, x2, y2)


def get_biggest_bbox(
    bbox: Tuple[float, float, float, float], new_bbox: Tuple[float, float, float, float]
) -> Tuple[float, float, float, float]:
    x1, y1, x2, y2 = bbox
    x1_new, y1_new, x2_new, y2_new = new_bbox
    return get_bbox(min(x1, x1_new), min(y1, y1_new), max(x2, x2_new), max(y2, y2_new))


def merge_polygons(polygons: Sequence[Tuple[float, ...]]) -> Tuple[float, ...]:
    all_polygons = [Polygon(zip(polygon[::2], polygon[1::2])) for polygon in polygons]
    merged_polygon = all_polygons[0]
    for polygon in all_polygons[1:]:
        merged_polygon = merged_polygon.union(polygon)
    # Check if the merged polygon is a MultiPolygon
    if isinstance(merged_polygon, MultiPolygon):
        # If it's a MultiPolygon, merge all the polygons into a single polygon
        merged_polygon = Polygon(merged_polygon.convex_hull)

    # Convert the merged polygon back to the format (x1, y1, x2, y2, ...)
    merged_coords = list(merged_polygon.exterior.coords)
    merged_coords = [coord for point in merged_polygon.exterior.coords for coord in point]
    return tuple(float(coord) for coord in merged_coords)


def iou_polygons(a: Tuple[float, ...], b: Tuple[float, ...]) -> float:
    # Convert tuples to numpy arrays
    a_array = np.array(a).reshape(-1, 2)
    b_array = np.array(b).reshape(-1, 2)

    # Create polygon objects
    poly_a = Polygon(a_array)
    poly_b = Polygon(b_array)

    # Ensure polygons are valid
    poly_a = make_valid(poly_a)
    poly_b = make_valid(poly_b)

    try:
        # Calculate intersection and union areas
        intersection_area = float(poly_a.intersection(poly_b).area)
        union_area = float(poly_a.area + poly_b.area - intersection_area)

        # Calculate IoU
        iou = intersection_area / union_area if union_area > 0 else 0.0

        return iou
    except (TopologicalError, ValueError) as e:
        print(f"Error calculating IoU: {e}. This may be due to invalid polygon geometries.")
        return 0.0
    except ZeroDivisionError:
        print("Error calculating IoU: Union area is zero. The polygons might be empty or invalid.")
        return 0.0
    except Exception as e:
        print(f"Error calculating IoU: {e}")
        return 0.0


def check_polygon_containment(
    polygon1: Tuple[float, ...], polygon2: Tuple[float, ...], threshold=1
):
    """
    Check if polygon1 contains polygon2 and calculate the containment percentage.

    :param polygon1: List of coordinates for polygon1 in format [x1,y1,x2,y2,x3,y3...]
    :param polygon2: List of coordinates for polygon2 in format [x1,y1,x2,y2,x3,y3...]
    :param threshold: Percentage threshold for considering containment (default 100%)
    :return: Tuple (is_contained, containment_percentage)
    """
    # Convert coordinate lists to list of tuples
    polygon1_points = [(polygon1[i], polygon1[i + 1]) for i in range(0, len(polygon1), 2)]
    polygon2_points = [(polygon2[i], polygon2[i + 1]) for i in range(0, len(polygon2), 2)]

    # Create Shapely Polygon objects
    polygon1 = Polygon(polygon1_points)
    polygon2 = Polygon(polygon2_points)

    # Calculate the intersection of the two polygons
    intersection = polygon1.intersection(polygon2)

    # Calculate the containment percentage
    containment_percentage = intersection.area / polygon2.area

    # Check if the containment percentage meets the threshold
    is_contained = containment_percentage >= threshold

    return is_contained, containment_percentage


def check_point_in_polygon(point: Tuple[float, float], polygon: Tuple[float, ...]) -> bool:
    """
    Check if a point lies inside a polygon.

    Args:
        point: Tuple of (x, y) coordinates
        polygon: Tuple of polygon coordinates in format (x1,y1,x2,y2,...)

    Returns:
        bool: True if point is inside polygon, False otherwise
    """
    # Convert polygon coordinates to list of tuples
    polygon_points = [(polygon[i], polygon[i + 1]) for i in range(0, len(polygon), 2)]

    # Create Shapely Point and Polygon objects
    point_obj = Point(point)
    polygon_obj = Polygon(polygon_points)

    # Check if point is inside or on the boundary of the polygon
    return polygon_obj.contains(point_obj) or polygon_obj.boundary.contains(point_obj)



import os
import random
import re

from typing import List

from pflows.typedef import Dataset, Image


def sample_group_proportion(dataset: Dataset, number: int, seed: int = 49) -> Dataset:
    shuffled_images = dataset.images
    random.seed(seed)
    random.shuffle(shuffled_images)
    groups_numbers = {}
    for image in shuffled_images:
        if image.group not in groups_numbers:
            groups_numbers[image.group] = 0
        groups_numbers[image.group] += 1
    # calculate the number of images to take per group
    for group in list(groups_numbers.keys()):
        groups_numbers[group] = int(number * (groups_numbers[group] / len(shuffled_images)))
    if sum(groups_numbers.values()) < number:
        # if the sum of the number of images is less than the number
        # we will add the remaining images to the group with the most images
        remaining = number - sum(groups_numbers.values())
        max_group = max(groups_numbers, key=groups_numbers.get)  # type: ignore
        groups_numbers[max_group] += remaining
    # take the images
    images = []
    for group in list(groups_numbers.keys()):
        images += [image for image in shuffled_images if image.group == group][
            : groups_numbers[group]
        ]

    return Dataset(
        images=images,
        categories=dataset.categories,
        groups=dataset.groups,
    )


def sample(
    dataset: Dataset, number: int, offset: int = 0, seed: int = 49, sort: str | None = None
) -> Dataset:
    if dataset.groups is not None and len(dataset.groups) > 1:
        if offset is not None and offset > 0:
            raise ValueError("Offset is not supported when groups are present")
        return sample_group_proportion(dataset, number, seed)

    sorted_images = dataset.images
    new_ids_order = [image.id for image in sorted_images]
    if sort is not None:
        new_ids_order = sorted(new_ids_order)
    else:
        # shuffle the images using seed
        random.seed(seed)
        random.shuffle(new_ids_order)

    new_images = []
    for new_id in new_ids_order:
        image = [image for image in sorted_images if image.id == new_id][0]
        new_images.append(image)
    return Dataset(
        images=new_images[offset : offset + number],
        categories=dataset.categories,
        groups=dataset.groups,
    )


def by_ids(dataset: Dataset, ids: list[str]) -> Dataset:
    return Dataset(
        images=[image for image in dataset.images if image.id in ids],
        categories=dataset.categories,
        groups=dataset.groups,
    )


def name_duplicate(dataset: Dataset, regexp: str) -> Dataset:
    # convert to regexp
    name_regexp_to_compare = re.compile(regexp)
    # We want to check for duplicates in the regexp name
    name_groups: dict[str, List[Image]] = {}
    for image in dataset.images:
        image_name = os.path.basename(image.path)
        match = re.match(name_regexp_to_compare, image_name)
        if match:
            name = match.group(0)
            if name not in name_groups:
                name_groups[name] = []
            name_groups[name].append(image)
    exclude_images = []
    for name, group_images in name_groups.items():
        if len(group_images) > 1:
            # We need to exclude all the other paths
            exclude_images += [image.path for index, image in enumerate(group_images) if index != 0]
    print(f"found {len(exclude_images)} duplicate images")
    dataset.images = [image for image in dataset.images if image.path not in exclude_images]
    return dataset


def by_group(dataset: Dataset, group: str) -> Dataset:
    return Dataset(
        images=[image for image in dataset.images if group == image.group],
        categories=dataset.categories,
        groups=dataset.groups,
    )


def by_groups(dataset: Dataset, groups: List[str]) -> Dataset:
    return Dataset(
        images=[image for image in dataset.images if image.group in groups],
        categories=dataset.categories,
        groups=dataset.groups,
    )


def by_category_name(
    dataset: Dataset, include: List[str] | None = None, exclude: List[str] | None = None
) -> Dataset:
    include = include or []
    exclude = exclude or []
    filtered_images = []

    for image in dataset.images:
        category_names = set(ann.category_name for ann in image.annotations)

        if exclude and any(cat in exclude for cat in category_names):
            continue

        if not include or any(cat in include for cat in category_names):
            filtered_images.append(image)

    return Dataset(images=filtered_images, categories=dataset.categories, groups=dataset.groups)



import os
import uuid
import tempfile
import mimetypes
from functools import wraps
from typing import Any, Dict, Tuple, Sequence, cast

import fastapi
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse
from fastapi.exceptions import HTTPException

from modal import App, asgi_app, Function, Image, gpu, Volume, Dict as ModalDict, Secret
from modal.functions import FunctionCall


from pflows.workflow import run_workflow

jobs_volume = Volume.from_name("pflows_jobs", create_if_missing=True)
uploads_volume = Volume.from_name("pflows_upload", create_if_missing=True)

persisted_jobs_dict: Dict[str, Any] = cast(
    Dict[str, Any], ModalDict.from_name("pflows_jobs_dict", create_if_missing=True)
)

GPU_TYPE = gpu.A100(size="40GB", count=1)

image = (
    Image.debian_slim()
    .apt_install("libgl1-mesa-dev")
    .apt_install("libglib2.0-0")
    .pip_install_from_requirements("./requirements.txt")
)

image_gpu = (
    Image.debian_slim()
    .apt_install(["ffmpeg", "libsm6", "libxext6"])
    .pip_install_from_requirements("./requirements.txt", gpu=GPU_TYPE)
)

app = App("pflows")

web_app = fastapi.FastAPI()


@app.function(
    image=image,
    volumes={"/root/jobs_data": jobs_volume, "/root/uploads_data": uploads_volume},
    secrets=[Secret.from_name("PFLOW_AUTH_TOKEN")],
    timeout=600,
)
@asgi_app()
def fastapi_app() -> Any:
    return web_app


def set_env(env_variables: Dict[str, Any], job_id: str) -> None:
    for key, value in env_variables.items():
        os.environ[key] = value
    if os.environ.get("BASE_FOLDER") is None:
        with tempfile.TemporaryDirectory() as tmp:
            os.environ["BASE_FOLDER"] = tmp
            os.environ["REMOTE_BASE_FOLDER"] = os.environ["BASE_FOLDER"]
            print(f"Setting BASE_FOLDER to {tmp}")
    os.environ["JOB_ID"] = job_id

    os.environ["PERSISTED_FOLDER"] = f"/root/jobs_data/{job_id}"
    os.environ["REMOTE_PERSISTED_FOLDER"] = os.environ["PERSISTED_FOLDER"]
    if not os.path.exists(os.environ["PERSISTED_FOLDER"]):
        os.makedirs(os.environ["PERSISTED_FOLDER"], exist_ok=True)
    os.environ["TEMP_FOLDER"] = tempfile.gettempdir()
    os.environ["REMOTE_TEMP_FOLDER"] = os.environ["TEMP_FOLDER"]


@app.function(
    image=image,
    timeout=7 * 60 * 60,
    volumes={"/root/jobs_data": jobs_volume, "/root/uploads_data": uploads_volume},
)
def endpoint_run_workflow_cpu(
    workflow: Sequence[Dict[str, Any]], env: Dict[str, Any], job_id: str
) -> Dict[str, Any]:
    print("Running workflow")
    set_env(env, job_id)
    results = run_workflow(
        raw_workflow=workflow, store_dict=persisted_jobs_dict, store_dict_key=job_id
    )
    persisted_jobs_dict[job_id] = {**results, "status": "completed"}
    return cast(Dict[str, Any], persisted_jobs_dict[job_id])


@app.function(
    image=image_gpu,
    gpu=GPU_TYPE,
    timeout=7 * 60 * 60,
    volumes={"/root/jobs_data": jobs_volume, "/root/uploads_data": uploads_volume},
)
def endpoint_run_workflow_gpu(
    workflow: Sequence[Dict[str, Any]], env: Dict[str, Any], job_id: str
) -> Dict[str, Any]:
    print("Running workflow")
    set_env(env, job_id)
    results = run_workflow(
        raw_workflow=workflow, store_dict=persisted_jobs_dict, store_dict_key=job_id
    )
    persisted_jobs_dict[job_id] = {**results, "status": "completed"}
    return cast(Dict[str, Any], persisted_jobs_dict[job_id])


def get_request(request_json: Dict[str, Any]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    workflow = request_json.get("workflow")
    if workflow is None:
        raise ValueError("The workflow is required.")
    env_variables = request_json.get("env") or {}
    return workflow, env_variables


def start_job(request_json: Dict[str, Any], endpoint: Function) -> Dict[str, Any]:
    job_id = uuid.uuid4().hex
    try:
        workflow, env = get_request(request_json)
    except ValueError as e:
        return {"error": str(e)}
    call = endpoint.spawn(workflow, env, job_id)
    if call is None:
        return {"error": "Failed to start workflow"}
    persisted_jobs_dict[job_id] = {"call_id": call.object_id, "status": "running"}
    return {"call_id": call.object_id, "job_id": job_id, "status": "running"}


def auth_required(endpoint_func: Any) -> Any:
    @wraps(endpoint_func)
    async def wrapper(*args: Any, **kwargs: Any) -> Any:
        equal_tokens = False
        try:
            request: fastapi.Request | None = (
                kwargs.get("request") or kwargs.get("_request") or None
            )
            if request is None:
                raise ValueError("Request not found")
            token = (request.headers.get("authorization") or "").split(" ")[1]
            equal_tokens = os.environ.get("PFLOW_AUTH_TOKEN") == token
        # pylint: disable=broad-exception-caught
        except Exception:
            pass

        if not equal_tokens:
            raise HTTPException(
                status_code=fastapi.status.HTTP_401_UNAUTHORIZED,
                detail="Incorrect bearer token",
                headers={"WWW-Authenticate": "Bearer"},
            )
        return await endpoint_func(*args, **kwargs)

    return wrapper


@web_app.post("/workflow/gpu")
@auth_required
async def workflow_gpu(request: fastapi.Request) -> Dict[str, Any]:
    request_json = await request.json()
    print("workflow_gpu")
    return start_job(request_json, endpoint_run_workflow_gpu)


@web_app.post("/workflow/cpu")
@auth_required
async def workflow_cpu(request: fastapi.Request) -> Dict[str, Any]:
    request_json = await request.json()
    print("workflow_cpu")
    return start_job(request_json, endpoint_run_workflow_cpu)


@web_app.get("/result/{job_id}")
@auth_required
async def poll_results(_request: fastapi.Request, job_id: str) -> Any:
    try:
        job_info = persisted_jobs_dict[job_id]
    except KeyError:
        return {"error": "Job not found."}
    if job_info.get("status") == "completed":
        return JSONResponse(content=jsonable_encoder(persisted_jobs_dict[job_id]))
    function_call = FunctionCall.from_id(job_info["call_id"])
    try:
        json_result = function_call.get(timeout=0)
        persisted_jobs_dict[job_id] = {
            **json_result,
            **persisted_jobs_dict[job_id],
            "status": "completed",
        }
        return JSONResponse(content=jsonable_encoder(persisted_jobs_dict[job_id]))

    except TimeoutError:
        http_accepted_code = 202
        return fastapi.responses.JSONResponse(
            content=jsonable_encoder(job_info), status_code=http_accepted_code
        )


@web_app.post("/download/{job_id}")
@auth_required
async def download_data(request: fastapi.Request, job_id: str) -> Any:
    request_path = (await request.json()).get("path")
    print("request_path", request_path)
    if request_path is None:
        return {"error": "The path is required."}
    request_path = request_path.lstrip("/")

    sanitized_path = os.path.normpath(request_path)

    if ".." in sanitized_path:
        return {"error": "Invalid path."}

    abs_path = f"/root/jobs_data/{job_id}/{request_path}"

    print(abs_path)
    if not os.path.exists(abs_path):
        return {"error": "File not found."}

    content_type, _ = mimetypes.guess_type(abs_path)
    if content_type is None:
        content_type = "application/octet-stream"  # Default content type if unknown

    # Read the file content
    with open(abs_path, "rb") as file:
        file_content = file.read()

    # Return the file content with the guessed content type
    return fastapi.Response(content=file_content, media_type=content_type)


@web_app.post("/uploads/download/{upload_id}")
@auth_required
async def download_upload_file(request: fastapi.Request, upload_id: str) -> Any:
    request_path = (await request.json()).get("path")
    print("request_path", request_path)
    if request_path is None:
        return {"error": "The path is required."}
    request_path = request_path.lstrip("/")

    sanitized_path = os.path.normpath(request_path)

    if ".." in sanitized_path:
        return {"error": "Invalid path."}

    abs_path = f"/root/uploads_data/{upload_id}/{request_path}"

    if not os.path.exists(abs_path):
        return {"error": "File not found."}

    content_type, _ = mimetypes.guess_type(abs_path)
    if content_type is None:
        content_type = "application/octet-stream"

    # Read the file content
    with open(abs_path, "rb") as file:
        file_content = file.read()

    # Return the file content with the guessed content type
    return fastapi.Response(content=file_content, media_type=content_type)


@web_app.post("/uploads")
@auth_required
async def upload_file(request: fastapi.Request) -> Any:
    data = await request.form()
    file = data["file"]
    path = data["path"]
    upload_id = uuid.uuid4().hex
    abs_path = f"/root/uploads_data/{upload_id}/{path}"
    os.makedirs(os.path.dirname(abs_path), exist_ok=True)
    if isinstance(file, str):
        with open(abs_path, "w", encoding="utf-8") as f:
            f.write(file)
    else:
        with open(abs_path, "wb") as f:
            f.write(await file.read())

    return {
        "status": "uploaded",
        "path": path,
        "upload_id": upload_id,
        "size": os.path.getsize(abs_path),
        "abs_path": abs_path,
    }


@web_app.post("/uploads/delete/{upload_id}")
@auth_required
async def delete_upload_file(request: fastapi.Request, upload_id: str) -> Any:
    request_path = (await request.json()).get("path") or ""
    if request_path is None:
        return {"error": "The path is required."}

    request_path = request_path.split("/")[-1]

    sanitized_path = os.path.normpath(request_path)

    if ".." in sanitized_path:
        return {"error": "Invalid path."}

    abs_path = f"/root/uploads_data/{upload_id}/{request_path}"

    print(abs_path)
    if not os.path.exists(abs_path):
        return {"error": "File not found."}

    os.remove(abs_path)
    return {"status": "deleted"}



import os
import json
import tempfile

from pflows.workflow import read_workflow


workflow_w_variable = [
  {
    "task": "set_env_var",
    "name": "DATA_ID",
    "value": "group3"
  },
  {
    "task": "yolo_v8.load_dataset",
    "folder_path": "{{BASE_FOLDER}}/datasets/downloaded/{{DATA_ID}}"
  }
]

def test_read_w_variable_path():
    with tempfile.TemporaryDirectory() as tmp:
        path = os.path.join(tmp, 'workflow_w_variable.json')
        with open(path, 'w') as f:
            f.write(json.dumps(workflow_w_variable))
        os.environ['BASE_FOLDER'] = tmp
        workflow, workflow_data = read_workflow(workflow_path=path)
        assert len(workflow) == 1
        first_task = workflow[0]
        assert first_task.params['folder_path'] == f"{tmp}/datasets/downloaded/group3"
  

def test_read_w_variable_raw_workflow():
    with tempfile.TemporaryDirectory() as tmp:
        os.environ['BASE_FOLDER'] = tmp
        workflow, workflow_data = read_workflow(raw_workflow=workflow_w_variable)
        assert len(workflow) == 1
        first_task = workflow[0]
        assert first_task.params['folder_path'] == f"{tmp}/datasets/downloaded/group3"



import os
import json
from typing import Dict, Any, cast

import requests


class PFlowsRemoteError(Exception):
    pass


def prepare_headers(token: str | None) -> Dict[str, str]:
    headers = {
        "Content-Type": "application/json",
    }
    if token:
        headers["Authorization"] = f"Bearer {token}"
    return headers


def run(
    base_url: str,
    token: str | None,
    workflow: str | Dict[str, Any],
    env: Dict[str, str] | None = None,
    gpu: bool = False,
) -> Dict[str, Any]:
    """
    Run a workflow on the remote server.
    :param base_url: The base URL of the remote server.
    :param token: The token to authenticate the request.
    :param workflow: The workflow to run. Can be the workflow_path as str or a dictionary.
    :param env: The environment variables to pass to the workflow.
    :param gpu: Whether to run the workflow on the GPU.
    :return: The response from the server."""

    headers = prepare_headers(token)
    mode = "gpu" if gpu else "cpu"
    url = f"{base_url}/workflow/{mode}"
    response = requests.post(
        url,
        headers=headers,
        data=json.dumps(
            {
                "workflow": workflow,
                "env": env or {},
            }
        ),
        timeout=60 * 60 * 2,
    )
    return cast(Dict[str, Any], response.json())


def download_job_file(
    base_url: str, token: str | None, job_id: str, remote_path: str, local_path: str
) -> None:
    headers = prepare_headers(token)

    url = f"{base_url}/download/{job_id}"
    response = requests.post(
        url,
        headers=headers,
        json={"path": remote_path},
        timeout=60 * 30,
    )

    if response.status_code == 200:
        content_type = response.headers.get("Content-Type") or ""

        target_dir = os.path.dirname(local_path)
        if target_dir != "":
            os.makedirs(target_dir, exist_ok=True)
        if content_type == "application/json":
            data = response.json()
            with open(local_path, "w", encoding="utf-8") as f:
                json.dump(data, f)
        elif content_type.startswith("text/"):
            with open(local_path, "w", encoding="utf-8") as f:
                f.write(response.text)
        else:
            with open(local_path, "wb") as f:
                f.write(response.content)
    else:
        raise PFlowsRemoteError(
            f"Failed to download file: {response.status_code} - {response.text}"
        )


def result(base_url: str, token: str | None, job_id: str) -> Dict[str, Any]:
    headers = prepare_headers(token)
    url = f"{base_url}/result/{job_id}"
    response = requests.get(url, headers=headers, timeout=60 * 10)
    return cast(Dict[str, Any], response.json())


def upload_file(
    base_url: str, token: str | None, local_path: str, remote_name: str
) -> Dict[str, Any]:
    headers = prepare_headers(token)
    del headers["Content-Type"]

    with open(local_path, "rb") as f:
        url = f"{base_url}/uploads"
        files = {"file": (remote_name, f)}
        data = {"path": remote_name}
        response = requests.post(url, headers=headers, data=data, files=files, timeout=60 * 60)
        if response.status_code != 200:
            raise PFlowsRemoteError(
                f"Failed to upload file: {response.status_code} - {response.text}"
            )
        return cast(Dict[str, Any], response.json())


def download_upload_file(
    base_url: str, token: str | None, upload_id: str, remote_path: str, local_path: str
) -> None:
    headers = prepare_headers(token)

    url = f"{base_url}/uploads/download/{upload_id}"
    response = requests.post(url, headers=headers, json={"path": remote_path}, timeout=60 * 30)

    if response.status_code == 200:
        content_type = response.headers.get("Content-Type") or ""

        os.makedirs(os.path.dirname(local_path), exist_ok=True)
        if content_type == "application/json":
            data = response.json()
            with open(local_path, "w", encoding="utf-8") as f:
                json.dump(data, f)
        elif content_type.startswith("text/"):
            with open(local_path, "w", encoding="utf-8") as f:
                f.write(response.text)
        else:
            with open(local_path, "wb") as f:
                f.write(response.content)
    else:
        raise PFlowsRemoteError(
            f"Failed to download file: {response.status_code} - {response.text}"
        )



import sys
import json
import os
import time
from pathlib import Path
from typing import Dict, Any, List

import matplotlib.pyplot as plt
import numpy as np
from dotenv import load_dotenv

from pflows import workflow
from pflows.tools import pflows_remote


# Load environment variables
def load_environment(override_file: str | None = None) -> Dict[str, Any]:
    """
    Load environment variables from a .env file specified as a command-line argument,
    with an optional override file provided as a parameter.
    Returns a dictionary with the loaded configuration.
    """
    load_dotenv()

    # Load the override file if provided
    if override_file:
        if os.path.exists(override_file):
            load_dotenv(override_file, override=True)
            print(f"Loaded override file: {override_file}")
        else:
            print(f"Warning: Override file {override_file} not found. Skipping.")

    required_vars = [
        "PFLOWS_BASE_URL",
        "PFLOWS_AUTH_TOKEN",
        "YOLO_MODEL",
        "EPOCHS",
        "DATA_FOLDER",
        "OUTPUT_MODEL_NAME",
    ]

    config = {}

    for var in required_vars:
        value = os.getenv(var)
        if value is None:
            raise ValueError(f"{var} must be provided")
        config[var] = value

    # Convert EPOCHS to int
    config["EPOCHS"] = int(config["EPOCHS"])
    config["OUTPUT_MODEL_NAME_PT"] = f"{config['OUTPUT_MODEL_NAME']}.pt"

    return config


# Upload dataset workflow
def create_upload_workflow(config: Dict[str, Any], zip_filepath: str) -> List[Dict[str, Any]]:
    return [
        {
            "task": "base.compress_folder",
            "output": zip_filepath,
            "compress_path": config["DATA_FOLDER"],
        },
        {
            "id": "upload_file",
            "task": "pflows_remote.upload_file",
            "base_url": config["PFLOWS_BASE_URL"],
            "token": config["PFLOWS_AUTH_TOKEN"],
            "local_path": zip_filepath,
            "remote_name": f"{Path(config['DATA_FOLDER']).name}.zip",
        },
    ]


# Remote training workflow
def create_remote_workflow(config: Dict[str, Any], upload_id: str) -> List[Dict[str, Any]]:
    return [
        {
            "task": "base.decompress_zip",
            "zip_path": f"/root/uploads_data/{upload_id}/{Path(config['DATA_FOLDER']).name}.zip",
            "output": "{{REMOTE_TEMP_FOLDER}}/dataset",
        },
        {
            "id": "train",
            "task": "yolo_v8.train",
            "model_name": f"{config['YOLO_MODEL']}.pt",
            "epochs": config["EPOCHS"],
            "data_file": "{{REMOTE_TEMP_FOLDER}}/dataset/data.yaml",
            "model_output": f"{{{{REMOTE_PERSISTED_FOLDER}}}}/models/{config['OUTPUT_MODEL_NAME_PT']}",
        },
    ]


# Process and save metrics
def process_metrics(job_results: Dict[str, Any], metrics_output_folder: str):
    response_json = job_results["train"]["results"]
    metrics = response_json["box"]
    precision, recall, f1_score = metrics["p"], metrics["r"], metrics["f1"]
    ap_class_index = metrics["ap_class_index"]

    class_names = response_json.get("names").values()
    ap_class_names = [response_json["names"][str(i)] for i in ap_class_index]

    aggregated_metrics = {
        "precision": sum(precision) / len(precision),
        "recall": sum(recall) / len(recall),
        "f1_score": sum(f1_score) / len(f1_score),
    }

    confusion_matrix = response_json["confusion_matrix"]["matrix"]

    class_metrics = {}
    for class_index, class_name in enumerate(class_names):
        try:
            ap_index = ap_class_names.index(class_name)
        except ValueError:
            continue
        class_metrics[class_name] = {
            "precision": round(precision[ap_index], 3),
            "recall": round(recall[ap_index], 3),
            "f1_score": round(f1_score[ap_index], 3),
            "n": int(
                sum(
                    [
                        confusion_matrix[row_index][class_index]
                        for row_index in range(len(confusion_matrix))
                    ]
                )
            ),
        }

    sorted_class_metrics = dict(
        sorted(class_metrics.items(), key=lambda x: x[1]["n"], reverse=True)
    )

    metrics_json = {
        "aggregated_metrics": aggregated_metrics,
        "class_metrics": sorted_class_metrics,
    }

    os.makedirs(metrics_output_folder, exist_ok=True)

    with open(f"{metrics_output_folder}/metrics.json", "w", encoding="utf-8") as f:
        json.dump(metrics_json, f, indent=4)

    confusion_matrix_json = {"matrix": confusion_matrix, "class_names": list(class_names)}
    with open(f"{metrics_output_folder}/confusion_matrix.json", "w", encoding="utf-8") as f:
        json.dump(confusion_matrix_json, f, indent=4)

    with open(f"{metrics_output_folder}/results.json", "w", encoding="utf-8") as f:
        json.dump(response_json, f, indent=4)

    # Generate confusion matrix plot
    plt.figure(figsize=(12, 10))
    plt.imshow(confusion_matrix, cmap="hot", interpolation="nearest")
    plt.colorbar(fraction=0.046, pad=0.04)
    axis_classes = list(class_names) + ["(background)"]
    plt.xticks(np.arange(len(axis_classes)), axis_classes, rotation=90)
    plt.yticks(np.arange(len(axis_classes)), axis_classes)
    plt.xlabel("True")
    plt.ylabel("Predicted")
    plt.title("Confusion Matrix")
    plt.tight_layout()
    plt.savefig(f"{metrics_output_folder}/confusion_matrix.png")

    print(f"Metrics saved to: {metrics_output_folder}")
    print(f"Confusion matrix image: {metrics_output_folder}/confusion_matrix.png")


def main(override_file: str | None = None):
    """
    Main function to run the train server.
    """
    config = {}
    if override_file is None:
        config = load_environment()
    else:
        config = load_environment(override_file)

    zip_filepath = f"dataset_{Path(config['DATA_FOLDER']).name}.zip"

    # Upload dataset
    upload_workflow = create_upload_workflow(config, zip_filepath)
    print(upload_workflow)
    upload_results = workflow.run_workflow(raw_workflow=upload_workflow)
    print(
        f"Upload details: {{'local_path': '{zip_filepath}', 'remote_name': '{Path(config['DATA_FOLDER']).name}.zip'}}"
    )
    print(upload_results)

    print("Waiting for the upload to finish")
    time.sleep(60)

    UPLOAD_ID = upload_results["upload_file"]["upload_id"]
    print(f"/root/uploads_data/{UPLOAD_ID}/{Path(config['DATA_FOLDER']).name}.zip")

    # Remote training
    remote_workflow = create_remote_workflow(config, UPLOAD_ID)
    train_output = pflows_remote.run(
        base_url=config["PFLOWS_BASE_URL"],
        token=config["PFLOWS_AUTH_TOKEN"],
        workflow=remote_workflow,
        env={"UPLOAD_ID": UPLOAD_ID},
        gpu=True,
    )
    job_id = train_output["job_id"]
    print("job_id", job_id)

    while True:
        job_results = pflows_remote.result(
            base_url=config["PFLOWS_BASE_URL"], token=config["PFLOWS_AUTH_TOKEN"], job_id=job_id
        )
        if job_results["status"] == "completed":
            break
        print(job_results)
        time.sleep(30)

    # Process metrics
    metrics_output_folder = f"metrics_{config['OUTPUT_MODEL_NAME']}"
    process_metrics(job_results, metrics_output_folder)

    print("Waiting to download the results")
    time.sleep(180)

    # Retrieve the model
    download_result = pflows_remote.download_job_file(
        base_url=config["PFLOWS_BASE_URL"],
        token=config["PFLOWS_AUTH_TOKEN"],
        job_id=job_id,
        remote_path=f"models/{config['OUTPUT_MODEL_NAME_PT']}",
        local_path=config["OUTPUT_MODEL_NAME_PT"],
    )

    print(download_result)
    print(f"Model saved to: {config['OUTPUT_MODEL_NAME_PT']}")


if __name__ == "__main__":
    main()



import os
import json
from datetime import datetime, date
from pathlib import Path
from hashlib import md5
import shutil
import tempfile
from typing import List, Tuple, Callable, Any, Dict
import zipfile

import torch
import yaml
from PIL import Image as ImagePil
from ultralytics import YOLO
import cv2
import numpy as np
from skimage.measure import approximate_polygon
from numpy.typing import NDArray

from pflows.tools.categories import remap_category_ids
from pflows.typedef import Annotation, Category, Dataset
from pflows.polygons import (
    calculate_center_from_bbox,
    calculate_center_from_polygon,
    bbox_from_polygon,
    polygon_from_bbox,
)
from pflows.model import get_image_info

GROUPS_ALIAS = {"val": "val", "test": "test", "valid": "val", "train": "train"}
ROUNDING = 6


def get_item_from_numpy_or_tensor(element: torch.Tensor | np.ndarray[Any, Any] | Any) -> Any:
    if isinstance(element, torch.Tensor):
        # element is a Tensor
        values = element.numpy().item()
    elif isinstance(element, np.ndarray):
        # element is a NumPy array
        values = element.item()
    else:
        # element is neither a Tensor nor a NumPy array
        values = element
    return values


def bbox_from_yolo_v8(
    polygon_row: Tuple[float, float, float, float]
) -> Tuple[float, float, float, float]:
    x_center, y_center, width, height = polygon_row
    return (
        round(x_center - width / 2, ROUNDING),
        round(y_center - height / 2, ROUNDING),
        round(x_center + width / 2, ROUNDING),
        round(y_center + height / 2, ROUNDING),
    )


def yolov8_from_bbox(bbox: Tuple[float, float, float, float]) -> Tuple[float, float, float, float]:
    x1, y1, x2, y2 = bbox
    width = x2 - x1
    height = y2 - y1
    x_center = x1 + width / 2
    y_center = y1 + height / 2
    return x_center, y_center, width, height


def process_image_annotations(
    label_path: str, categories: List[Category], mode: str = "auto"
) -> List[Annotation]:
    annotations = []
    with open(label_path, "r", encoding="utf-8") as f:
        for index, line in enumerate(f.readlines()):
            rows = line.split(" ")
            category_id = int(rows[0])
            polygon_row = [round(float(x), ROUNDING) for x in rows[1:]]
            # Generate a unique id for the annotation
            data = f"{label_path}_{category_id}_{index}_{', '.join(str(x) for x in polygon_row)}"
            md5_hash = md5(data.encode()).hexdigest()
            bbox = None
            polygon = None
            obb = None
            candidate_task = None
            if len(polygon_row) == 4:
                bbox_row_tuple = (
                    polygon_row[0],
                    polygon_row[1],
                    polygon_row[2],
                    polygon_row[3],
                )
                bbox = bbox_from_yolo_v8(bbox_row_tuple)
                polygon = polygon_from_bbox(bbox)
                task = "detect"
            else:
                if len(polygon_row) == 8:
                    task = "obb"
                    obb = (
                        polygon_row[0],
                        polygon_row[1],
                        polygon_row[2],
                        polygon_row[3],
                        polygon_row[4],
                        polygon_row[5],
                        polygon_row[6],
                        polygon_row[7],
                    )
                else:
                    task = "segment"
                polygon = tuple(polygon_row)
                bbox = bbox_from_polygon(polygon)
            if mode == "bbox":
                task = "detect"
            elif mode == "segment":
                task = "segment"

            annotations.append(
                Annotation(
                    id=md5_hash,
                    category_id=category_id,
                    category_name=categories[category_id].name,
                    center=calculate_center_from_polygon(polygon),
                    bbox=bbox,
                    segmentation=polygon,
                    task=task,
                    obb=obb,
                )
            )
    return annotations


def get_categories_from_model(model):
    categories = []
    if isinstance(model.names, dict):
        for key in get_model_category_ids(model):
            categories.append(model.names[key])
    else:
        categories = model.names
    return categories


def get_model_category_ids(model):
    model_names_keys = sorted(
        model.names.keys() if isinstance(model.names, dict) else range(len(model.names))
    )
    return model_names_keys


def load_categories(parsed_yaml_file) -> List[Category]:
    classes = parsed_yaml_file["names"]
    if isinstance(classes, dict):
        classes = list(classes.values())
    return [Category(name=str(cls), id=index) for index, cls in enumerate(classes)]


def load_dataset(
    dataset: Dataset | None, folder_path: str, number: int | None = None, mode: str = "auto"
) -> Dataset:
    if dataset is None:
        dataset = Dataset(images=[], categories=[], groups=[])
    print()
    print("Loading dataset from yolo_v8 format in: ", folder_path)
    data_yaml = Path(folder_path) / "data.yaml"
    if not os.path.exists(data_yaml):
        raise ValueError(f"File {data_yaml} does not exist")
    with open(data_yaml, "r", encoding="utf-8") as f:
        data = yaml.load(f, Loader=yaml.FullLoader)
        categories = load_categories(data)
        groups = {
            GROUPS_ALIAS[file]: str((Path(folder_path) / file).resolve())
            for file in os.listdir(folder_path)
            if file in GROUPS_ALIAS
        }
        images = []
        reach_max = False
        for group_name, group_folder in groups.items():
            if reach_max:
                break
            if not os.path.exists(group_folder):
                raise ValueError(f"Group {group_name} does not exist")

            images_folder = Path(group_folder) / "images"
            for image_path in os.listdir(images_folder):
                if number is not None and len(images) >= number:
                    reach_max = True
                    break
                image_target_path = images_folder / image_path
                if not os.path.exists(image_target_path):
                    raise ValueError(f"Image {image_target_path} does not exist")
                image_info = get_image_info(str(image_target_path), group_name)
                if not os.path.exists(image_info.path):
                    image_info.path = str(image_target_path)
                image_info.annotations = process_image_annotations(
                    str(Path(group_folder) / "labels" / (image_target_path.stem + ".txt")),
                    categories,
                    mode,
                )
                images.append(image_info)
        category_names = list(set([category.name for category in categories + dataset.categories]))
        new_groups = list(set(list(groups.keys()) + dataset.groups))
        return remap_category_ids(
            Dataset(
                images=images + dataset.images,
                categories=[
                    Category(name=name, id=index) for index, name in enumerate(category_names)
                ],
                groups=new_groups,
            )
        )


def preprocess_image(
    image: ImagePil.Image,
    new_size: Tuple[int, int] = (640, 640),
    grayscale: bool = False,
) -> NDArray[np.uint8]:
    # 1. Stretch to 640x640
    new_image = image.resize(new_size)

    # Convert to numpy array for OpenCV processing
    image_np = np.array(new_image)

    # Check the number of channels
    if len(image_np.shape) == 2 or (len(image_np.shape) == 3 and image_np.shape[2] == 1):
        # Image is grayscale
        if not grayscale:
            # Convert to RGB if color output is desired
            image_np = cv2.cvtColor(image_np, cv2.COLOR_GRAY2RGB)
    elif len(image_np.shape) == 3:
        if image_np.shape[2] == 4:
            # Image is RGBA, convert to RGB
            image_np = cv2.cvtColor(image_np, cv2.COLOR_RGBA2RGB)
        elif image_np.shape[2] == 3:
            # Image is RGB, no conversion needed
            pass
        else:
            raise ValueError(f"Unexpected number of channels: {image_np.shape[2]}")

        if grayscale:
            # Convert to grayscale if grayscale output is desired
            image_np = cv2.cvtColor(image_np, cv2.COLOR_RGB2GRAY)
    else:
        raise ValueError(f"Unexpected image shape: {image_np.shape}")

    if grayscale:
        clahe = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8, 8))
        image_np = clahe.apply(image_np)
        if len(image_np.shape) == 2:
            image_np = cv2.cvtColor(image_np, cv2.COLOR_GRAY2RGB)
    else:
        lab = cv2.cvtColor(image_np, cv2.COLOR_RGB2LAB)
        l, a, b = cv2.split(lab)
        clahe = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8, 8))
        lab = cv2.merge((clahe.apply(l), a, b))
        image_np = cv2.cvtColor(lab, cv2.COLOR_LAB2RGB)

    return image_np.astype(np.uint8)


# work in the pylint errors
# pylint: disable=too-many-arguments,too-many-locals
def run_model_on_image(
    image_path: str,
    model: YOLO,
    model_categories: List[str] | None = None,
    threshold: float = 0.5,
    segment_tolerance: float = 0,
    preprocess_function: Callable[[ImagePil.Image], NDArray[np.uint8]] = preprocess_image,
    add_tag: str | None = None,
    model_id: str | None = None,
) -> List[Annotation]:
    # We run the model on the image
    image = ImagePil.open(image_path)

    results = model.predict(
        preprocess_function(image),
        conf=threshold,
    )
    annotations = []
    yolo_categories = model.names
    if isinstance(yolo_categories, dict):
        yolo_categories = [str(yolo_categories[key]) for key in sorted(yolo_categories.keys())]
    if model_categories is None:
        model_categories = yolo_categories
    task = model.task
    print(task)
    if not task in ["detect", "segment", "obb"]:
        raise ValueError(f"Task {task} is not supported")
    task = str(task)
    for result in results:
        # Segmentation case
        if result.masks is not None and result.boxes is not None:
            for np_mask, box in zip(result.masks.xyn, result.boxes):
                category_id = int(box.cls[0])
                category_name = model_categories[category_id]

                points = np_mask.ravel().tolist()
                polygon_array = np.array(points).reshape(-1, 2)
                # We didn't got good results with the approximate_polygon
                # simplified_polygon = approximate_polygon(
                #     polygon_array, tolerance=segment_tolerance
                # )  # type: ignore[no-untyped-call]
                # simplified_points = tuple(
                #     round(x, ROUNDING) for x in simplified_polygon.flatten().tolist()
                # )
                simplified_points = tuple(
                    round(x, ROUNDING) for x in polygon_array.flatten().tolist()
                )
                if len(simplified_points) < 4:
                    continue

                joined_points = ", ".join(str(x) for x in simplified_points)
                hash_id = md5(
                    f"{image_path}_{category_id}_{box.conf[0]}_{joined_points}".encode()
                ).hexdigest()
                annotations.append(
                    Annotation(
                        id=hash_id,
                        category_id=category_id,
                        category_name=category_name,
                        center=calculate_center_from_polygon(simplified_points),
                        bbox=bbox_from_polygon(simplified_points),
                        segmentation=simplified_points,
                        task=task,
                        conf=round(get_item_from_numpy_or_tensor(box.conf[0]), ROUNDING),
                        tags=[add_tag] if add_tag else [],
                        model_id=model_id,
                    )
                )
            continue
        # Classification case
        if result.boxes is None and result.probs is not None:
            for index, prob in enumerate(result.probs.numpy().tolist()):
                # we get the index of the class with the highest probability
                if prob < threshold:
                    continue
                category_name = model_categories[index]
                annotations.append(
                    Annotation(
                        id=md5(f"{image_path}_{index}_{prob}".encode()).hexdigest(),
                        category_id=index,
                        category_name=category_name,
                        center=None,
                        bbox=None,
                        segmentation=None,
                        task="classification",
                        conf=round(prob, ROUNDING),
                        tags=[add_tag] if add_tag else [],
                        model_id=model_id,
                    )
                )
            continue

        # Bounding box case
        if result.boxes is not None:
            for box in result.boxes:
                bbox = tuple(round(x, ROUNDING) for x in box.xyxyn[0].tolist())
                category_id = int(get_item_from_numpy_or_tensor(box.cls))
                category_name = model_categories[category_id]
                confidence = round(get_item_from_numpy_or_tensor(box.conf), ROUNDING)
                hash_id = md5(
                    (
                        f"{image_path}_{category_id}_{confidence}_{', '.join(str(x) for x in bbox)}"
                    ).encode()
                ).hexdigest()
                annotations.append(
                    Annotation(
                        id=hash_id,
                        category_id=category_id,
                        category_name=category_name,
                        center=calculate_center_from_bbox(bbox),
                        bbox=bbox,
                        segmentation=polygon_from_bbox(bbox),
                        task="detect",
                        conf=confidence,
                        tags=[add_tag] if add_tag else [],
                        model_id=model_id,
                    )
                )
        if result.obb is not None:
            for obb in result.obb:
                category_id = int(get_item_from_numpy_or_tensor(obb.cls))
                category_name = model_categories[category_id]
                confidence = round(get_item_from_numpy_or_tensor(obb.conf), ROUNDING)

                # Extract the OBB normalized coordinates
                obb_coords = obb.xyxyxyxyn.tolist()[0]  # Normalized coordinates

                # Calculate the center of the OBB (normalized)
                center_x = sum(x for x, _ in obb_coords) / 4
                center_y = sum(y for _, y in obb_coords) / 4
                center = (round(center_x, ROUNDING), round(center_y, ROUNDING))

                # Calculate an approximate bounding box (normalized)
                x_coords, y_coords = zip(*obb_coords)
                bbox = (
                    round(min(x_coords), ROUNDING),
                    round(min(y_coords), ROUNDING),
                    round(max(x_coords), ROUNDING),
                    round(max(y_coords), ROUNDING),
                )

                # Round the segmentation coordinates
                segmentation_pairs = [
                    (round(x, ROUNDING), round(y, ROUNDING)) for x, y in obb_coords
                ]
                segmentation_flat = [item for sublist in segmentation_pairs for item in sublist]

                hash_id = md5(
                    f"{image_path}_{category_id}_{confidence}_{','.join(map(str, segmentation_flat))}".encode()
                ).hexdigest()

                annotations.append(
                    Annotation(
                        id=hash_id,
                        category_id=category_id,
                        category_name=category_name,
                        center=center,
                        bbox=bbox,
                        segmentation=segmentation_flat,
                        obb=segmentation_flat,
                        task="obb",
                        conf=confidence,
                        tags=[add_tag] if add_tag else [],
                        model_id=model_id,
                    )
                )
    return sorted(
        annotations,
        key=lambda x: x.conf,
        reverse=True,
    )


def performance_non_max_suppression(boxes, scores, iou_threshold):
    """
    Perform non-maximum suppression (NMS) on the bounding boxes.

    Args:
        boxes (numpy.ndarray): Array of bounding boxes, each represented as [x1, y1, x2, y2].
        scores (numpy.ndarray): Array of corresponding confidence scores for each box.
        iou_threshold (float): Intersection over Union (IoU) threshold for
        suppressing overlapping boxes.

    Returns:
        list: Indices of the selected boxes after applying NMS.
    """
    # Sort the boxes by their confidence scores in descending order
    sorted_indices = np.argsort(scores)[::-1]

    selected_indices = []

    while len(sorted_indices) > 0:
        # Select the box with the highest confidence score
        current_index = sorted_indices[0]
        selected_indices.append(current_index)

        # Compute the IoU between the current box and the remaining boxes
        current_box = boxes[current_index]
        remaining_indices = sorted_indices[1:]
        remaining_boxes = boxes[remaining_indices]

        # Compute the coordinates of the intersection area
        x1 = np.maximum(current_box[0], remaining_boxes[:, 0])
        y1 = np.maximum(current_box[1], remaining_boxes[:, 1])
        x2 = np.minimum(current_box[2], remaining_boxes[:, 2])
        y2 = np.minimum(current_box[3], remaining_boxes[:, 3])

        # Compute the area of the intersection
        intersection_area = np.maximum(0, x2 - x1) * np.maximum(0, y2 - y1)

        # Compute the area of the current box and the remaining boxes
        current_area = (current_box[2] - current_box[0]) * (current_box[3] - current_box[1])
        remaining_areas = (remaining_boxes[:, 2] - remaining_boxes[:, 0]) * (
            remaining_boxes[:, 3] - remaining_boxes[:, 1]
        )

        # Compute the IoU
        iou = intersection_area / (current_area + remaining_areas - intersection_area)

        # Remove the boxes with IoU greater than the threshold
        keep_indices = np.where(iou <= iou_threshold)[0]
        sorted_indices = remaining_indices[keep_indices]

    return selected_indices


def annotations_non_max_suppression(
    annotations: List[Annotation],
    categories=None,
    iou_threshold: float = 0.4,
    active: bool = True,
) -> List[Annotation]:
    keep_annotations = []
    if categories is None:
        categories = [annotation.category_id for annotation in annotations]
    for category_id in categories:
        category_annotations = [
            annotation for annotation in annotations if annotation.category_id == category_id
        ]
        if not category_annotations:
            continue
        if active:
            keep_annotations_indexes = performance_non_max_suppression(
                np.array([annotation.bbox for annotation in category_annotations]),
                np.array([annotation.conf for annotation in category_annotations]),
                iou_threshold,
            )
            keep_annotations.extend(
                [category_annotations[index] for index in keep_annotations_indexes]
            )
        else:
            keep_annotations.extend(category_annotations)
    return keep_annotations


def run_model(
    dataset: Dataset,
    model_path: str,
    add_tag: str | None = None,
    threshold: float = 0.5,
    segment_tolerance: float = 0.02,
    non_max_suppression: bool = True,
    non_max_suppression_threshold: float = 0.4,
) -> Dataset:
    model = YOLO(model_path)
    model_names_keys = get_model_category_ids(model)
    model_names = [model.names[key] for key in sorted(model_names_keys)]
    current_category_names = [category.name for category in dataset.categories]
    new_categories = current_category_names + [
        name for name in model_names if name not in current_category_names
    ]
    print("New categories found: ", new_categories)
    total_images = len(dataset.images)
    for index, image in enumerate(dataset.images):
        model_annotations = run_model_on_image(
            image.path,
            model,
            threshold=threshold,
            segment_tolerance=segment_tolerance,
            add_tag=add_tag,
        )
        keep_annotations = annotations_non_max_suppression(
            model_annotations,
            model_names_keys,
            iou_threshold=non_max_suppression_threshold,
            active=non_max_suppression,
        )

        image.annotations = image.annotations + keep_annotations
        if index % 25 == 0:
            progress_percentage = round((index / total_images) * 100, 2)
            print()
            print("-" * 50)
            print(f"Progress: {progress_percentage}%")
            print("-" * 50)
            print()

    dataset.categories = [
        Category(name=name, id=index) for index, name in enumerate(new_categories)
    ]
    return remap_category_ids(dataset)


# TODO: refactor
# pylint: disable=too-many-branches,too-many-statements
def write(
    dataset: Dataset,
    target_dir: str,
    pre_process_images: bool = False,
    remove_existing: bool = False,
    write_original: bool = True,
) -> None:
    raw_path = Path(target_dir) / "raw"
    target_path = Path(target_dir) / "yolo"
    data_yaml_path = target_path / "data.yaml"
    if target_path.exists() and remove_existing:
        if data_yaml_path.exists():
            # remove the directory
            shutil.rmtree(target_path)
            if raw_path.exists():
                shutil.rmtree(raw_path)
        else:
            raise ValueError(
                f"Directory {target_path} already exists, but is not a yolo_v8 dataset"
            )

    target_path.mkdir(parents=True, exist_ok=True)
    data_for_yaml = {
        "names": [category.name for category in dataset.categories],
        "nc": len(dataset.categories),
    }
    for group in dataset.groups:
        data_for_yaml[group] = f"./{group}/images"
    with open(data_yaml_path, "w", encoding="utf-8") as f:
        yaml.dump(
            data_for_yaml,
            f,
        )
    total_images: dict[str, int] = {}
    for image in dataset.images:
        # Copy the original image
        raw_folder = raw_path / image.group
        if write_original and pre_process_images:
            image_folder = raw_folder / "images"
            image_folder.mkdir(parents=True, exist_ok=True)
            image_path = image_folder / f"{image.id}.jpg"
            shutil.copy(image.path, image_path)

        # write image information
        dataset_path = raw_path / image.group / "dataset"
        dataset_path.mkdir(parents=True, exist_ok=True)
        with open(dataset_path / f"{image.id}.json", "w", encoding="utf-8") as f:
            f.write(json.dumps(image.__dict__, cls=CustomEncoder, indent=4))

        # write confidences
        confidences_folder = target_path / image.group / "confidences"
        confidences_folder.mkdir(parents=True, exist_ok=True)
        confidence_path = confidences_folder / f"{image.id}.txt"
        with open(confidence_path, "w", encoding="utf-8") as f:
            for annotation in image.annotations:
                f.write(f"{annotation.conf or ''}\n")

        # write image information
        image_info_folder = target_path / image.group / "image_info"
        image_info_folder.mkdir(parents=True, exist_ok=True)
        image_info_path = image_info_folder / f"{image.id}.json"
        with open(image_info_path, "w", encoding="utf-8") as f:
            f.write(json.dumps(image.info, cls=CustomEncoder, indent=4))

        image_folder = target_path / image.group / "images"
        image_folder.mkdir(parents=True, exist_ok=True)
        image_path = image_folder / f"{image.id}.jpg"
        with ImagePil.open(image.path) as img:
            if pre_process_images:
                new_image = preprocess_image(img)
                ImagePil.fromarray(new_image).save(image_path)  # type: ignore
            else:
                img.save(image_path)
        total_images[image.group] = total_images.get(image.group, 0) + 1
        label_folder = target_path / image.group / "labels"
        label_folder.mkdir(parents=True, exist_ok=True)
        label_path = label_folder / f"{image.id}.txt"
        with open(label_path, "w", encoding="utf-8") as f:
            for annotation in image.annotations:
                if annotation.task == "detect" and annotation.bbox:
                    yolo_bbox = yolov8_from_bbox(annotation.bbox)
                    f.write(f"{annotation.category_id} {' '.join(str(x) for x in yolo_bbox)}\n")
                elif annotation.task in "segment" and annotation.segmentation:
                    polygon_str = " ".join(str(x) for x in annotation.segmentation)
                    f.write(f"{annotation.category_id} {polygon_str}\n")
                elif annotation.task == "obb" and annotation.obb:
                    obb_str = " ".join(str(x) for x in annotation.obb)
                    f.write(f"{annotation.category_id} {obb_str}\n")
                elif annotation.task == "obb" and annotation.segmentation:
                    polygon_str = " ".join(str(x) for x in annotation.segmentation)
                    f.write(f"{annotation.category_id} {polygon_str}\n")
    print()
    print("Dataset saved as yolo_v8 format in: ", target_dir)
    print("total images: ", len(dataset.images))
    print("total images per group:")
    for group, total in total_images.items():
        print(f"\t{group}: {total}")


def check_device() -> str:
    if torch.backends.mps.is_available():
        device = "mps"
    else:
        device = "cuda" if torch.cuda.is_available() else "cpu"
    return device


class CustomEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, np.ndarray):
            return o.tolist()
        if isinstance(o, (datetime, date)):
            return o.isoformat()
        if isinstance(o, Path):
            return str(o)
        try:
            return o.__dict__
        except AttributeError:
            pass
        return super().default(o)


def train(
    dataset: Dataset,
    data_file: str,
    model_name: str,
    model_output: str,
    epochs: int = 100,
    batch_size: int = 8,
) -> Dict[str, Any]:
    model = YOLO(model_name)
    device = check_device()
    print("Training on device: ", device)

    results = model.train(
        data=data_file,
        epochs=epochs,
        batch=batch_size,
        imgsz=640,
        device=device,
        val=True,
    )
    if results is None:
        print("Training failed")
        return {"dataset": dataset}
    try:
        results_dict = {
            key: value for key, value in results.__dict__.items() if not key.startswith("on_")
        }
    # pylint: disable=broad-except
    except Exception:
        results_dict = {
            "box": json.loads(json.dumps(results.box.__dict__ or {}, cls=CustomEncoder)),
            "seg": json.loads(json.dumps(results.seg.__dict__ or {}, cls=CustomEncoder)),
        }

    # Save the model
    current_model_location = Path(results.save_dir) / "weights" / "best.pt"
    model_output_dir = Path(model_output).parent
    model_output_dir.mkdir(parents=True, exist_ok=True)
    shutil.move(current_model_location, model_output)
    print("Model saved in: ", model_output)
    # results dict and ignore the key starting with on_

    return {
        "dataset": dataset,
        "results": json.loads(json.dumps(results_dict, cls=CustomEncoder)),
        "model_output": model_output,
    }


def find_data_yaml_folder(temp_dir: str) -> str | None:
    for root, _, files in os.walk(temp_dir):
        for file in files:
            if file.endswith("data.yaml"):
                return root
    return None


def load_from_zip(zip_path: str, temp_dir: str | None = None) -> Dataset:
    if temp_dir is None:
        temp_dir = tempfile.mkdtemp()
    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        zip_ref.extractall(temp_dir)
    # find data.yaml in the zip inside the temp_dir, recursively
    data_yaml_folder_path = find_data_yaml_folder(temp_dir)
    if not data_yaml_folder_path:
        raise ValueError("data.yaml not found in the zip file")
    return load_dataset(None, data_yaml_folder_path)



from hashlib import md5
from dataclasses import asdict
from typing import Any, Sequence

from pflows.polygons import (
    bbox_from_polygon,
    calculate_center_from_polygon,
    merge_polygons,
    polygon_from_bbox,
)
from pflows.typedef import Annotation, Dataset, Category, Image


def keep(dataset: Dataset, categories: list[str]) -> Dataset:
    """
    Filters a dataset by keeping only the specified categories.

    This function takes a dataset and a list of category names and returns a new dataset
    containing only the images and annotations that belong to the specified categories.
    It iterates over each image in the dataset and filters its annotations based on the
    provided categories. If an image has at least one annotation belonging to the specified
    categories, it is included in the new dataset along with its filtered annotations.

    Args:
        dataset: The input dataset to be filtered.
        categories: A list of category names to keep in the filtered dataset.

    Returns:
        A new dataset containing only the images and annotations belonging to the
        specified categories.

    Example:
        # Define the categories to keep
        categories_to_keep = [
            "11", "12", "13", "14", "15", "16", "17", "18"
        ]

        # Call the keep function
        filtered_dataset = keep(dataset, categories_to_keep)

        # The filtered_dataset will only contain images and annotations belonging to the
        # specified categories.
        # For example, if an image has annotations with categories "11", "12", and "33", only
        # the annotations with categories "11" and "12" will be kept, while the annotation with
        # category "33" will be removed.
        # If an image has no annotations belonging to the specified categories, it will be excluded
        # from the filtered dataset.
    """
    new_categories = [category for category in dataset.categories if category.name in categories]
    print()
    print("New categories: ")
    for category in new_categories:
        print("\t", category.name)
    return remap_category_ids(
        Dataset(
            images=[
                Image(
                    **{
                        **asdict(image),
                        "annotations": [
                            annotation
                            for annotation in image.annotations
                            if annotation.category_name in categories
                        ],
                    }
                )
                for image in dataset.images
                if any(annotation.category_name in categories for annotation in image.annotations)
            ],
            categories=new_categories,
            groups=dataset.groups,
        )
    )


def group_categories(
    dataset: Dataset, groups: dict[str, Sequence[Sequence[str]]], condition: str = "any"
) -> Dataset:
    """
    Groups annotations in a dataset based on the provided category groups.

    Args:
        dataset: The input dataset containing images and annotations.
        groups: A dictionary specifying the category groups. The keys represent the new category
                names, and the values are lists of lists, where each inner list contains the
                categories to be grouped together.
        condition: "any" or "all". If "any", the annotations will be grouped if they have at least
                one category in common. If "all", the annotations will be grouped only if they
                have all categories in common. Default is "any".

    Returns:
        A new dataset with the annotations grouped according to the specified category groups.

    Raises:
        ValueError: If the annotations being grouped have a task other than "detect".

    Example:
        # Define the category groups
        groups = {
            "11_31": [["11", "31"], ["21", "41"]],
            "12_32": [["12", "32"], ["22", "42"]],
            "11": [["11"]],
            "12": [["12"]],
            ...
        }

        # Call the group_categories function
        new_dataset = group_categories(dataset, groups)

        # The new_dataset will contain annotations grouped according to the specified groups
        # For example, annotations with categories "11" and "31" will be merged into a single
        # annotation with the new category "11_31". Similarly, annotations with categories
        # "12" and "32" will be merged into a single annotation with the new category "12_32".
        # The size of the bbox with be the size of the biggest bbox among the annotations
        # in the group.
    """
    new_images = []
    categories_names = list(groups.keys())
    for image in dataset.images:
        new_annotations = []
        all_replace_categories = [el for rows in groups.values() for row in rows for el in row]
        task = [
            ann.task for ann in image.annotations if ann.category_name in all_replace_categories
        ][0]
        for new_category, replace_category_groups in groups.items():
            polygon = None
            for replace_category_group in replace_category_groups:
                annotations = [
                    annotation
                    for annotation in image.annotations
                    if annotation.category_name in replace_category_group
                ]
                if len(annotations) == 0:
                    continue

                if condition == "all" and len(
                    list(set(annotation.category_name for annotation in annotations))
                ) != len(replace_category_group):
                    continue

                annotations_polygons = [
                    annotation.segmentation
                    for annotation in annotations
                    if annotation.segmentation is not None
                ]
                if len(annotations_polygons) == 0:
                    annotations_polygons = [
                        polygon_from_bbox(annotation.bbox)
                        for annotation in annotations
                        if annotation.bbox is not None
                    ]
                    if len(annotations_polygons) == 0:
                        continue
                new_polygon = merge_polygons(annotations_polygons)
                polygon = merge_polygons([p for p in [polygon, new_polygon] if p is not None])

            if polygon is None:
                continue
            bbox = bbox_from_polygon(polygon)
            new_annotations.append(
                Annotation(
                    id=md5(
                        "_".join([str(annotation.id) for annotation in annotations]).encode()
                    ).hexdigest(),
                    category_id=categories_names.index(new_category),
                    center=calculate_center_from_polygon(polygon),
                    bbox=bbox,
                    segmentation=polygon,
                    task=image.annotations[0].task,
                    conf=(
                        round(
                            sum(annotation.conf for annotation in annotations) / len(annotations),
                            3,
                        )
                        if len(annotations) > 0
                        else -1.0
                    ),
                    category_name=new_category,
                    tags=list(set(tag for annotation in annotations for tag in annotation.tags)),
                )
            )

        new_images.append(
            Image(
                **{
                    **asdict(image),
                    "annotations": new_annotations,
                }
            )
        )
    return remap_category_ids(
        Dataset(
            images=new_images,
            categories=[
                Category(
                    id=index,
                    name=name,
                )
                for index, name in enumerate(categories_names)
            ],
            groups=dataset.groups,
        )
    )


def remap_category_ids(
    dataset: Dataset, check_missing_categories=True, dataclasses: dict[str, Any] | None = None
) -> Dataset:
    """
    Remaps the category ids in a dataset to be contiguous integers starting from 0.
    """
    if dataclasses is None:
        dataclasses = {}

    Image_class = dataclasses.get("Image", Image)
    Annotation_class = dataclasses.get("Annotation", Annotation)
    Category_class = dataclasses.get("Category", Category)
    Dataset_class = dataclasses.get("Dataset", Dataset)

    category_names = sorted(list(set(category.name for category in dataset.categories)))
    if check_missing_categories:
        category_names = sorted(
            list(
                set(
                    annotation.category_name
                    for image in dataset.images
                    for annotation in image.annotations
                )
            )
        )

    new_categories = [
        Category_class(
            id=index,
            name=name,
        )
        for index, name in enumerate(category_names)
    ]

    # Remap the category ids
    category_map = {category.name: category.id for category in new_categories}
    new_images = [
        Image_class(
            **{
                **asdict(image),
                "annotations": [
                    Annotation_class(
                        **{
                            **asdict(annotation),
                            "category_id": category_map[annotation.category_name],
                        }
                    )
                    for annotation in image.annotations
                ],
            }
        )
        for image in dataset.images
    ]
    return Dataset_class(
        images=new_images,
        categories=new_categories,
        groups=dataset.groups,
    )


def remap_categories(dataset: Dataset, mapping: dict[str, str]) -> Dataset:
    """
    Renames categories in a dataset based on the provided category map.

    Args:
        dataset: The input dataset containing images and annotations.
        category_map: A dictionary specifying the mapping of old category names to new category names.

    Returns:
        A new dataset with the categories renamed according to the specified category map.

    Example:
        # Define the category map
        category_map = {
            "11": "A",
            "12": "B",
            "13": "C",
            ...
        }

        # Call the rename_categories function
        new_dataset = rename_categories(dataset, category_map)

        # The new_dataset will contain categories renamed according to the specified category map.
        # For example, categories "11", "12", and "13" will be renamed to "A", "B", and "C"
        # respectively.
    """
    new_category_names = list(set(mapping.values()))
    new_categories = [
        Category(
            id=index,
            name=name,
        )
        for index, name in enumerate(new_category_names)
    ]
    return remap_category_ids(
        Dataset(
            images=[
                Image(
                    **{
                        **asdict(image),
                        "annotations": [
                            Annotation(
                                **{
                                    **asdict(annotation),
                                    "category_name": mapping.get(
                                        annotation.category_name, annotation.category_name
                                    ),
                                }
                            )
                            for annotation in image.annotations
                        ],
                    }
                )
                for image in dataset.images
            ],
            categories=new_categories,
            groups=dataset.groups,
        )
    )


def rename_some_categories(dataset: Dataset, renames: dict[str, str]) -> Dataset:
    """
    Renames specified categories in a dataset.

    This function takes a dataset and a dictionary of renames where keys are the original category names
    and values are the new names. It returns a new dataset with updated category names for the specified
    categories. Categories not specified in the renames dictionary remain unchanged.

    Args:
        dataset: The input dataset to be modified.
        renames: A dictionary mapping from old category names to new category names.

    Returns:
        A new dataset with updated category names.

    Example:
        # Define the renames
        renames = {
            "11": "A",
            "12": "B",
            "13": "C"
        }

        # Call the rename_some_categories function
        updated_dataset = rename_some_categories(dataset, renames)

        # The updated_dataset will have categories "11", "12", and "13" renamed to "A", "B", and "C"
        # respectively, while other categories remain unchanged.
    """
    new_images = [
        Image(
            **{
                **asdict(image),
                "annotations": [
                    Annotation(
                        **{
                            **asdict(annotation),
                            "category_name": renames.get(
                                annotation.category_name, annotation.category_name
                            ),
                        }
                    )
                    for annotation in image.annotations
                ],
            }
        )
        for image in dataset.images
    ]
    updated_categories = [
        Category(id=category.id, name=renames.get(category.name, category.name))
        for category in dataset.categories
    ]
    return remap_category_ids(
        Dataset(
            images=new_images,
            categories=updated_categories,
            groups=dataset.groups,
        )
    )



# pylint: disable=R0902

import random
from typing import Any, Callable, Dict, List, Optional, Tuple
from dataclasses import asdict, dataclass, field, replace

from PIL import Image as PILImage, ImageDraw


def generate_random_color() -> Tuple[int, int, int]:
    """Generate a random RGB color."""
    return (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))


def relative_to_absolute_coords(coords: Tuple[float, ...], width: int, height: int) -> List[int]:
    """
    Convert relative coordinates to absolute coordinates based on image dimensions.

    Args:
        coords (Tuple[float, ...]): Relative coordinates.
        width (int): Image width.
        height (int): Image height.

    Returns:
        List[int]: Absolute coordinates.
    """
    return [
        int(coords[i] * width if i % 2 == 0 else coords[i] * height) for i in range(len(coords))
    ]


@dataclass
class Category:
    """Represents a category for annotations."""

    id: int
    name: str


@dataclass
class Annotation:
    """Represents an annotation in an image."""

    id: str
    category_id: int
    center: (
        Tuple[float, float] | None
    )  # Format: (x, y) this is relative to the image size (between 0 and 1)
    bbox: (
        Tuple[float, float, float, float] | None
    )  # Format: (x1, y1, x2, y2) this is relative to the image size (between 0 and 1)
    segmentation: (
        Tuple[float, ...] | None
    )  # Format: (x1, y1, x2, y2, ...) this is relative to the image size (between 0 and 1)
    conf: float = -1.0  # Confidence score, between 0 and 1 (default: -1.0)
    category_name: str = ""
    tags: List[str] = field(default_factory=list)
    original_id: Optional[str] = None
    truncated: Optional[bool] = False
    model_id: Optional[str] = None
    obb: Optional[Tuple[float, ...]] = None  # Format (x1, y1, x2, y2, x3, y3, x4, y4)
    task: str = "detect"  # "segment" or "detect" or "obb"

    def distance(self, other_annotation) -> float:
        """
        Calculate the Euclidean distance between the centers of two annotations.

        Args:
            other_annotation (Annotation): Another annotation to compare with.

        Returns:
            float: Distance between the centers, or -1.0 if centers are not available.
        """
        if self.center and other_annotation.center:
            return (
                (self.center[0] - other_annotation.center[0]) ** 2
                + (self.center[1] - other_annotation.center[1]) ** 2
            ) ** 0.5
        return -1.0

    def dict(self) -> Dict[str, Any]:
        """Convert the annotation to a dictionary."""
        return asdict(self)

    @classmethod
    def from_dict(cls, data):
        """
        Create an Annotation instance from a dictionary, including fields from parent classes.

        Args:
            data (dict): Dictionary containing annotation data.

        Returns:
            Annotation: An instance of Annotation or its subclass.
        """
        if isinstance(data, cls):
            return data

        # Get all fields from the current class and its parent classes
        valid_fields = set()
        for c in cls.__mro__:
            if hasattr(c, "__annotations__"):
                valid_fields.update(c.__annotations__.keys())

        # Filter the input data to only include valid fields
        filtered_data = {k: v for k, v in data.items() if k in valid_fields}

        return cls(**filtered_data)


@dataclass
class Image:
    """Represents an image with annotations."""

    id: str
    path: str
    intermediate_ids: List[str]
    width: int
    height: int
    size_kb: int
    group: str
    annotations: List[Annotation] = field(default_factory=list)
    tags: List[str] = field(default_factory=list)
    info: Dict[str, Any] = field(default_factory=dict)

    def copy(self):
        """
        Create a shallow copy of the Image instance, including deep copies of mutable fields.

        Returns:
            Image: A copy of the Image instance.
        """
        new_image = replace(self)
        new_image.intermediate_ids = self.intermediate_ids.copy()
        new_image.annotations = [replace(ann) for ann in self.annotations]
        new_image.tags = self.tags.copy()
        new_image.info = self.info.copy()
        return new_image

    def draw(self) -> PILImage.Image:
        """
        Draw annotations on the image with:
        - Segmentation: semi-transparent fill
        - Detection: dotted border only (no fill)
        """
        img = PILImage.open(self.path)
        img = img.convert("RGB")
        overlay = PILImage.new("RGBA", img.size, (0, 0, 0, 0))
        draw = ImageDraw.Draw(overlay)
        width, height = img.size

        for annotation in self.annotations:
            # Generate color similar to web version
            hue = (annotation.category_id * 137.508) % 360
            h = hue / 360
            rgb_color = self._hsl_to_rgb(h, 0.7, 0.5)

            if annotation.task == "segment" and annotation.segmentation:
                # Segmentation gets fill and outline
                fill_color = (*rgb_color, 51)  # 20% opacity
                stroke_color = (*rgb_color, 25)  # 10% opacity

                abs_segmentation = relative_to_absolute_coords(
                    annotation.segmentation, width, height
                )
                points = list(zip(abs_segmentation[0::2], abs_segmentation[1::2]))
                draw.polygon(points, fill=fill_color, outline=stroke_color)

            elif annotation.task == "detect" and annotation.bbox:
                # Detection gets only dotted outline, no fill
                stroke_color = (*rgb_color, 255)  # Full opacity for border
                x1, y1, x2, y2 = annotation.bbox
                abs_coords = [int(x1 * width), int(y1 * height), int(x2 * width), int(y2 * height)]

                # Draw dotted rectangle (manually since PIL doesn't support dotted lines)
                dash_length = 5
                for i in range(0, int(x2 * width - x1 * width), dash_length * 2):
                    # Top line
                    draw.line(
                        [
                            abs_coords[0] + i,
                            abs_coords[1],
                            min(abs_coords[0] + i + dash_length, abs_coords[2]),
                            abs_coords[1],
                        ],
                        fill=stroke_color,
                    )
                    # Bottom line
                    draw.line(
                        [
                            abs_coords[0] + i,
                            abs_coords[3],
                            min(abs_coords[0] + i + dash_length, abs_coords[2]),
                            abs_coords[3],
                        ],
                        fill=stroke_color,
                    )

                for i in range(0, int(y2 * height - y1 * height), dash_length * 2):
                    # Left line
                    draw.line(
                        [
                            abs_coords[0],
                            abs_coords[1] + i,
                            abs_coords[0],
                            min(abs_coords[1] + i + dash_length, abs_coords[3]),
                        ],
                        fill=stroke_color,
                    )
                    # Right line
                    draw.line(
                        [
                            abs_coords[2],
                            abs_coords[1] + i,
                            abs_coords[2],
                            min(abs_coords[1] + i + dash_length, abs_coords[3]),
                        ],
                        fill=stroke_color,
                    )

            # Draw text with background
            if annotation.category_name:
                if annotation.bbox:
                    text_x = int(annotation.bbox[0] * width)
                    text_y = int(annotation.bbox[1] * height) - 15
                elif annotation.segmentation:
                    # Calculate text position for segmentation
                    x_coords = annotation.segmentation[0::2]
                    y_coords = annotation.segmentation[1::2]
                    text_x = int(min(x_coords) * width)
                    text_y = int(min(y_coords) * height) - 15
                else:
                    continue

                text = annotation.category_name
                if annotation.conf >= 0:
                    text += f" ({int(annotation.conf * 100)}%)"

                # Draw text background
                text_bbox = draw.textbbox((text_x, text_y), text)
                draw.rectangle(text_bbox, fill=(0, 0, 0, 178))  # 70% opacity black
                draw.text((text_x, text_y), text, fill=(255, 255, 255, 255))

        # Composite the transparent overlay onto the original image
        img = PILImage.alpha_composite(img.convert("RGBA"), overlay)
        return img.convert("RGB")

    def _hsl_to_rgb(self, h, s, l):
        """
        Convert HSL to RGB color.
        h, s, l values are in range [0, 1]
        Returns RGB values in range [0, 255]
        """

        def hue_to_rgb(p, q, t):
            if t < 0:
                t += 1
            if t > 1:
                t -= 1
            if t < 1 / 6:
                return p + (q - p) * 6 * t
            if t < 1 / 2:
                return q
            if t < 2 / 3:
                return p + (q - p) * (2 / 3 - t) * 6
            return p

        if s == 0:
            r = g = b = l
        else:
            q = l * (1 + s) if l < 0.5 else l + s - l * s
            p = 2 * l - q
            r = hue_to_rgb(p, q, h + 1 / 3)
            g = hue_to_rgb(p, q, h)
            b = hue_to_rgb(p, q, h - 1 / 3)

        return (int(r * 255), int(g * 255), int(b * 255))

    def dict(self) -> Dict[str, Any]:
        """Convert the image to a dictionary, including annotations."""
        return {**asdict(self), "annotations": [ann.dict() for ann in self.annotations]}

    @classmethod
    def from_dict(cls, data):
        """
        Create an Image instance from a dictionary, including fields from parent classes.

        Args:
            data (dict): Dictionary containing image data.

        Returns:
            Image: An instance of Image or its subclass.
        """
        if isinstance(data, cls):
            return data

        valid_fields = set()
        for c in cls.__mro__:
            if hasattr(c, "__annotations__"):
                valid_fields.update(c.__annotations__.keys())

        filtered_data = {k: v for k, v in data.items() if k in valid_fields}

        annotations = [Annotation.from_dict(ann) for ann in filtered_data.get("annotations", [])]
        filtered_data["annotations"] = annotations

        return cls(**filtered_data)


@dataclass
class Dataset:
    """Represents a dataset containing images, categories, and groups."""

    images: List[Image]
    categories: List[Category]
    groups: List[str]

    @classmethod
    def from_dict(cls, data):
        """
        Create a Dataset instance from a dictionary, discarding any fields not in the schema.

        Args:
            data (dict): Dictionary containing dataset data.

        Returns:
            Dataset: An instance of Dataset.
        """
        if isinstance(data, cls):
            return data

        valid_fields = set()
        for c in cls.__mro__:
            if hasattr(c, "__annotations__"):
                valid_fields.update(c.__annotations__.keys())

        filtered_data = {k: v for k, v in data.items() if k in valid_fields}

        images = [Image.from_dict(img) for img in filtered_data.get("images", [])]
        categories = [Category(**cat) for cat in filtered_data.get("categories", [])]
        groups = filtered_data.get("groups", [])

        return cls(images=images, categories=categories, groups=groups)


@dataclass
class Task:
    """Represents a task to be performed."""

    task: str
    function: Callable[..., Any]
    params: Dict[str, Any]
    skip: bool = False
    id: str | None = None


@dataclass
class Model:
    """Represents a model to be used."""

    id: str
    name: str
    task: str
    type: str
    categories: List[str]
    params: Dict[str, Any]
    version: str
    size_kb: int



import os
import time
import sys
import glob
from typing import Any, Dict, List
from pathlib import Path
import shutil
import zipfile


from pflows.model import get_image_info
from pflows.typedef import Dataset, Image

ALLOWED_IMAGES = [".jpg", ".png", ".jpeg"]


def count_images(dataset: Dataset) -> Dict[str, int]:
    print()
    print("total images: ", len(dataset.images))
    return {"count": len(dataset.images)}


def count_categories(dataset: Dataset) -> Dict[str, Any]:
    print()
    print("total categories: ", len(dataset.categories))
    # We count from annotations
    total_categories = {}
    categories_names = {}
    for image in dataset.images:
        for annotation in image.annotations:
            if annotation.category_id in total_categories:
                total_categories[annotation.category_id] += 1
            else:
                categories_names[annotation.category_id] = annotation.category_name
                total_categories[annotation.category_id] = 1
    print("Total Categories in annotations: ", len(total_categories))
    for category_id, total in total_categories.items():
        print("\t", category_id, categories_names[category_id], ":", total)
    return {"count": len(dataset.categories), "total_categories": total_categories}


def show_categories(dataset: Dataset) -> None:
    print()
    print("Categories:")
    for category in dataset.categories:
        print("\t", category.name)


def count_groups(dataset: Dataset) -> Dict[str, Any]:
    print()
    print("total groups: ", len(dataset.groups))
    groups_total = {}
    for group in dataset.groups:
        groups_total[group] = len([image for image in dataset.images if image.group == group])
        print("\t", group, ":", groups_total[group])
    return {"count": len(dataset.groups), "groups_total": groups_total}


def check_folder(folder: str) -> None:
    # We check if the folder exists and if its a folder
    if not os.path.exists(folder):
        raise FileNotFoundError("The folder does not exist")
    if not os.path.isdir(folder):
        raise NotADirectoryError("The specified path is not a directory")


def find_images_recursively(base_path: str) -> List[str]:
    # Patterns to match
    file_patterns = [f"*{ext}" for ext in ALLOWED_IMAGES]

    # List to hold all found file paths
    found_files = []

    # Recursively search for files
    for pattern in file_patterns:
        found_files.extend(glob.glob(os.path.join(base_path, "**", pattern), recursive=True))

    return found_files


def search_images_in_folder(folder: str, recursive: bool = False) -> List[str]:
    check_folder(folder)
    # We get all the images in the folder
    base_folder = Path(folder).resolve()
    images_paths = []
    if recursive:
        images_paths = find_images_recursively(str(base_folder))
    else:
        images_paths = [str((base_folder / file).resolve()) for file in os.listdir(base_folder)]
    return images_paths


def read_images_from_folder(folder: str, recursive: bool = False) -> List[Image]:
    check_folder(folder)
    # We get all the images in the folder
    base_folder = Path(folder).resolve()
    images: List[Image] = []
    images_paths = search_images_in_folder(folder, recursive=recursive)
    for image_path in images_paths:
        image_info = get_image_info(image_path, "train")
        images.append(image_info)
    return images


def load_images(
    dataset: Dataset,
    path: str,
    paths: List[str] | None = None,
    recursive: bool = False,
    number: int | None = None,
) -> Dataset:
    # we are going to load the images from the folder
    print()
    paths = paths or [path]
    print("loading images from:", paths)
    print("recursive:", recursive)

    image_paths = []
    if number is not None:
        print("number of images to load:", number)
    for folder_path in paths:
        if not os.path.exists(folder_path):
            continue
        if os.path.isfile(folder_path):
            image_paths.append(folder_path)
            continue
        image_paths += search_images_in_folder(folder_path, recursive=recursive)

    images: List[Image] = []

    if number is not None:
        image_paths = image_paths[:number]

    for image_path in image_paths:
        image_info = get_image_info(image_path, "train")
        images.append(image_info)

    print("total images loaded:", len(images))

    # remove duplicates ids
    already_seen = set()
    total_images = len(images)

    new_images = []
    for image in images:
        if image.id in already_seen:
            print("duplicated id", image.id)
            continue
        already_seen.add(image.id)
        new_images.append(image)

    print("removed duplicates ids on load", total_images - len(images))
    groups = ["train"]
    dataset.images += new_images
    if "train" not in dataset.groups:
        dataset.groups += groups
    return dataset


def terminate() -> None:
    sys.exit(0)


def wait(minutes: int | None = None, seconds: int | None = None) -> None:
    if minutes is not None and seconds is not None:
        raise ValueError("You can only specify one of the arguments")
    total_seconds = 0
    if minutes is not None:
        total_seconds += minutes * 60
    if seconds is not None:
        total_seconds += seconds
    print(f"Waiting for {total_seconds} seconds")
    time.sleep(total_seconds)


def echo(text: str, path: str = "") -> Dict[str, str]:
    if path:
        with open(path, "w", encoding="utf-8") as file:
            file.write(text)
    else:
        print(text)
    return {"echo": text}


def compress_folder(compress_path: str, output: str) -> Dict[str, str]:
    if not os.path.exists(compress_path):
        raise FileNotFoundError("The folder does not exist")
    clean_output = output
    if output.endswith(".zip"):
        clean_output = output[:-4]
    zip_path = f"{clean_output}.zip"
    if os.path.exists(zip_path):
        # remove the existing zip file
        os.remove(zip_path)
    shutil.make_archive(clean_output, "zip", compress_path)
    return {"status": "compressed", "compress_path": compress_path, "output": output}


def decompress_zip(zip_path: str, output: str) -> Dict[str, str]:
    if not os.path.exists(output):
        os.makedirs(output, exist_ok=True)

    print("decompressing zip file", zip_path, "to", output)

    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        zip_ref.extractall(output)
    return {"status": "uncompressed", "zip_path": zip_path, "output": output}


def print_sample(dataset: Dataset) -> None:
    print("Dataset Info:")
    print("Total images:", len(dataset.images))
    print("Total categories:", len(dataset.categories))
    print("Total groups:", len(dataset.groups))
    print("sample", dataset.images[0])



import os
import json
from typing import Dict, List, Optional, Any
from io import StringIO
import csv
from tabulate import tabulate


def find_metrics_files(root_folder: str, name_filter: Optional[str] = None) -> List[str]:
    metrics_files = []
    for root, _, files in os.walk(root_folder):
        for file in files:
            if file == "metrics.json":
                if name_filter is None or name_filter in root:
                    metrics_files.append(os.path.join(root, file))
    return metrics_files


def read_metrics_file(file_path: str) -> Dict[str, Any]:
    with open(file_path, "r", encoding="utf-8") as f:
        to_return: Dict[str, Any] = json.load(f)
        return to_return


def extract_metrics(data: Dict[str, Any]) -> Dict[str, float]:
    aggregated = data["aggregated_metrics"]
    return {
        "precision": aggregated["precision"],
        "recall": aggregated["recall"],
        "f1_score": aggregated["f1_score"],
    }


def create_comparison_table(
    metrics_data: Dict[str, Dict[str, float]],
    output_format: str = "markdown",
    sort_by: Optional[str] = None,
) -> str:
    headers = ["Folder", "Precision", "Recall", "F1 Score"]
    table_data = [
        [folder, metrics["precision"], metrics["recall"], metrics["f1_score"]]
        for folder, metrics in metrics_data.items()
    ]

    if sort_by:
        sort_index = ["precision", "recall", "f1_score"].index(sort_by) + 1
        table_data.sort(key=lambda x: x[sort_index], reverse=True)

    if output_format == "markdown":
        return tabulate(table_data, headers=headers, tablefmt="pipe", floatfmt=".4f")
    if output_format in ["csv", "tsv"]:
        delimiter = "," if output_format == "csv" else "\t"
        output = StringIO()
        writer = csv.writer(output, delimiter=delimiter)
        writer.writerow(headers)
        writer.writerows(table_data)
        return output.getvalue()
    raise ValueError(f"Unsupported format: {output_format}")


def show_detailed_metrics(
    data: Dict[str, Any],
    model_name: str,
    output_format: str = "markdown",
    sort_by: Optional[str] = None,
) -> str:
    aggregated = data["aggregated_metrics"]
    class_metrics = data["class_metrics"]

    agg_data = [
        ["Precision", aggregated["precision"]],
        ["Recall", aggregated["recall"]],
        ["F1 Score", aggregated["f1_score"]],
    ]

    class_data = [
        [class_name, metrics["precision"], metrics["recall"], metrics["f1_score"], metrics["n"]]
        for class_name, metrics in class_metrics.items()
    ]

    if sort_by:
        sort_index = ["precision", "recall", "f1_score"].index(sort_by) + 1
        class_data.sort(key=lambda x: x[sort_index], reverse=True)

    if output_format == "markdown":
        agg_table = tabulate(agg_data, headers=["Metric", "Value"], tablefmt="pipe", floatfmt=".4f")
        class_table = tabulate(
            class_data,
            headers=["Class", "Precision", "Recall", "F1 Score", "Count"],
            tablefmt="pipe",
            floatfmt=".4f",
        )
        return (
            f"Detailed Metrics for {model_name}\n\n"
            f"Aggregated Metrics:\n{agg_table}\n\n"
            f"Class-specific Metrics:\n{class_table}"
        )
    if output_format in ["csv", "tsv"]:
        delimiter = "," if output_format == "csv" else "\t"
        output = StringIO()
        writer = csv.writer(output, delimiter=delimiter)
        writer.writerow(["Metric", "Value"])
        writer.writerows(agg_data)
        writer.writerow([])  # Empty row as separator
        writer.writerow(["Class", "Precision", "Recall", "F1 Score", "Count"])
        writer.writerows(class_data)
        return output.getvalue()
    raise ValueError(f"Unsupported format: {output_format}")


def compare_models(
    root_folder: str,
    name_filter: Optional[str] = None,
    output_format: str = "markdown",
    sort_by: Optional[str] = None,
) -> str:
    metrics_files = find_metrics_files(root_folder, name_filter)
    metrics_data: Dict[str, Dict[str, float]] = {}

    for file_path in metrics_files:
        folder = os.path.dirname(file_path)
        data = read_metrics_file(file_path)
        metrics = extract_metrics(data)
        metrics_data[folder] = metrics

    return create_comparison_table(metrics_data, output_format, sort_by)


def find_metrics_file(model_path: str) -> Optional[str]:
    if os.path.isfile(model_path):
        model_dir = os.path.dirname(model_path)
    else:
        model_dir = model_path

    for root, _, files in os.walk(model_dir):
        if "metrics.json" in files:
            return os.path.join(root, "metrics.json")

    return None


def show_model_details(
    model_path: str, output_format: str = "markdown", sort_by: Optional[str] = None
) -> str:
    metrics_file = find_metrics_file(model_path)

    if not metrics_file:
        return f"Error: metrics.json not found for model {model_path}"

    data = read_metrics_file(metrics_file)
    return show_detailed_metrics(data, model_path, output_format, sort_by)



def run_function_path(dataset, file_path, function_name):
    import importlib.util
    import sys

    try:
        # Load the module from the given file path
        spec = importlib.util.spec_from_file_location("module.name", file_path)
        if spec is None:
            raise ImportError(f"Could not load module specification from {file_path}")
        if spec.loader is None:
            raise ImportError(f"Could not load module loader from {file_path}")

        module = importlib.util.module_from_spec(spec)
        sys.modules["module.name"] = module
        spec.loader.exec_module(module)

        # Get the function from the module
        function_to_run = getattr(module, function_name)

        # Execute the function with the provided dataset
        result = function_to_run(dataset)
        return result
    except ImportError as e:
        print(f"Error importing module: {e}")
    except AttributeError:
        print(f"Error: Function '{function_name}' not found in the module.")
    except Exception as e:
        print(f"An unexpected error occurred: {e}")

    return None  # Return None if any error occurs


def run_function(dataset, function):
    return function(dataset)



import os
import json
from hashlib import md5
from pathlib import Path
from typing import List

import imagesize

from pflows.typedef import Annotation, Image


def get_image_info_fast(image_path: str):
    # Get file size without reading the entire file
    size_bytes = os.path.getsize(image_path)
    size_kb = int(round(size_bytes / 1024, 2))

    # Open the image and get dimensions without loading the entire image into memory
    width, height = imagesize.get(image_path)

    # Compute MD5 hash of the file without loading it entirely into memory
    hasher = md5()
    with open(image_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hasher.update(chunk)
    image_hash = hasher.hexdigest()

    return width, height, size_kb, image_hash


def get_image_info(
    image_path: str, group_name: str, intermediate_ids: List[str] | None = None
) -> Image:
    width, height, size_kb, image_hash = get_image_info_fast(image_path)

    intermediate_ids = intermediate_ids or []
    image_path_obj = Path(image_path)
    if image_path_obj.name not in intermediate_ids:
        intermediate_ids.append(image_path_obj.name)
    if image_path_obj.stem not in intermediate_ids and image_path_obj.suffix in [
        "jpg",
        "jpeg",
        "png",
    ]:
        intermediate_ids.append(image_path_obj.stem)
    # check if raw folder exists
    image_stem = image_path_obj.stem
    raw_info_path = Path(
        str(image_path_obj)
        .replace("/yolo", "/raw")
        .replace(f"/images/{image_stem}", f"/dataset/{image_stem}")
    )
    raw_info_path = raw_info_path.with_suffix(".json")
    if raw_info_path.exists():
        return Image.from_dict(json.loads(raw_info_path.read_text()))

    image: Image = Image(
        id=image_hash,
        intermediate_ids=intermediate_ids,
        path=str(image_path),
        width=width,
        height=height,
        size_kb=size_kb,
        group=group_name,
    )
    return image


def image_from_image_path(
    path: str,
    annotations: List[Annotation] | None = None,
) -> Image:
    image = get_image_info(path, "train")
    image.annotations = annotations or []

    return image



from pathlib import Path
from typing import Dict, Any, List
import json

from pflows.tools.yolo_v8 import load_dataset, run_model
from pflows.tools.compare_datasets import compare_images_annotations
from pflows.tools.filter_images import by_group, by_groups
from pflows.typedef import Dataset, Image


def run_and_compare(
    model_path: str,
    dataset_path: str,
    groups: str | List[str] | None = None,
    threshold: float = 0.5,
    iou_threshold: float = 0.5,
    output_metrics: str | None = None,
) -> Dict[str, Any]:
    """
    Run YOLO model on images and compare with gold standard annotations.

    Args:
        model_path: Path to the YOLO model (.pt file)
        dataset_path: Path to the YOLO dataset folder
        groups: Single group name or list of group names to filter by (e.g. "train", ["train", "val"])
        threshold: Confidence threshold for model predictions
        iou_threshold: IoU threshold for matching predictions with ground truth
        output_metrics: Optional path to save metrics JSON file

    Returns:
        Dictionary containing comparison metrics
    """
    # Load gold standard dataset
    gold_dataset = load_dataset(None, dataset_path)

    # Filter by groups if specified
    if groups is not None:
        if isinstance(groups, str):
            gold_dataset = by_group(gold_dataset, groups)
        else:
            gold_dataset = by_groups(gold_dataset, groups)

    new_dataset_to_infer = Dataset(
        images=[
            Image(**{**image.copy().__dict__, "annotations": []}) for image in gold_dataset.images
        ],
        categories=gold_dataset.categories,
        groups=gold_dataset.groups,
    )
    # Run model on images
    inferred_dataset = run_model(
        dataset=new_dataset_to_infer,
        model_path=model_path,
        threshold=threshold,
        add_tag="inference",
    )
    # keep only

    # Get category names from gold dataset
    category_names = [category.name for category in gold_dataset.categories]

    # Compare results
    metrics = compare_images_annotations(
        gold_dataset.images, inferred_dataset.images, category_names, iou_threshold=iou_threshold
    )

    if output_metrics:
        # Save metrics to file
        metrics_dir = Path(output_metrics).parent
        metrics_dir.mkdir(parents=True, exist_ok=True)

        with open(output_metrics, "w", encoding="utf-8") as f:
            json.dump(metrics, f, indent=4)

        print(f"Metrics saved to: {output_metrics}")

    return metrics


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Run YOLO model and compare with gold standard")
    parser.add_argument("model_path", help="Path to YOLO model (.pt file)")
    parser.add_argument("dataset_path", help="Path to YOLO dataset folder")
    parser.add_argument("--groups", nargs="+", help="Groups to evaluate (e.g. train val)")
    parser.add_argument("--threshold", type=float, default=0.5, help="Confidence threshold")
    parser.add_argument("--iou-threshold", type=float, default=0.5, help="IoU threshold")
    parser.add_argument("--output", help="Path to save metrics JSON file")

    args = parser.parse_args()

    metrics = run_and_compare(
        args.model_path,
        args.dataset_path,
        args.groups,
        args.threshold,
        args.iou_threshold,
        args.output,
    )

    # Print summary metrics
    print("\nOverall Metrics:")
    print(f"Precision: {metrics['overall']['precision']:.4f}")
    print(f"Recall: {metrics['overall']['recall']:.4f}")
    print(f"F1 Score: {metrics['overall']['f1_score']:.4f}")

    print("\nMetrics by Category:")
    for category, cat_metrics in metrics["categories"].items():
        print(f"\n{category}:")
        print(f"  Precision: {cat_metrics['precision']:.4f}")
        print(f"  Recall: {cat_metrics['recall']:.4f}")
        print(f"  F1 Score: {cat_metrics['f1_score']:.4f}")



import json
import math
from pathlib import Path
import random
from typing import Any, Dict, List, Tuple
from pflows.typedef import Annotation, Category, Dataset, Image


# pylint: disable=too-many-locals
def re_split_dataset(
    dataset: Dataset,
    train_percentage: float = 0.7,
    val_percentage: float = 0.2,
    test_percentage: float = 0.1,
) -> Dict[str, Any]:
    # Validate percentages
    total_percentage = train_percentage + val_percentage + test_percentage
    rounded_total = math.ceil(total_percentage * 100)
    if rounded_total != 100:
        raise ValueError("The sum of the percentages must be 1.0")
    # Separate augmented images
    images_to_resplit = [image for image in dataset.images if "augmented" not in (image.tags or [])]
    augmented_images = [image for image in dataset.images if "augmented" in (image.tags or [])]
    random.shuffle(images_to_resplit)

    total_images = len(dataset.images)
    train_total_images, val_total_images, test_total_images = calculate_total_images(
        total_images, train_percentage, val_percentage, test_percentage, len(augmented_images)
    )
    groups_info = [
        ("train", 0, train_total_images),
        ("val", train_total_images, train_total_images + val_total_images),
        ("test", train_total_images + val_total_images, total_images),
    ]
    images = assign_groups(images_to_resplit, augmented_images, groups_info)

    # Determine groups
    groups = []
    if train_total_images > 0:
        groups.append("train")
    if val_total_images > 0:
        groups.append("val")
    if test_total_images > 0:
        groups.append("test")

    images_ids_by_group: Dict[str, List[Dict[str, Any]]] = {group: [] for group in groups}
    for image in images:
        images_ids_by_group[image.group].append(
            {
                "id": image.id,
                "name": Path(image.path).name,
                "intermediate_ids": image.intermediate_ids,
            }
        )
    return {
        "dataset": Dataset(images=images, categories=dataset.categories, groups=groups),
        "ids_by_group": images_ids_by_group,
    }


def calculate_total_images(
    total_images: int,
    train_percentage: float,
    val_percentage: float,
    test_percentage: float,
    num_augmented_images: int,
) -> Tuple[int, int, int]:
    train_total_images = int(total_images * train_percentage)
    train_total_images_without_augmented = train_total_images - num_augmented_images
    if train_total_images_without_augmented < 0:
        train_total_images = num_augmented_images
    else:
        train_total_images = train_total_images_without_augmented

    remaining_images = total_images - train_total_images
    val_total_images = int(remaining_images * (val_percentage / (val_percentage + test_percentage)))
    test_total_images = remaining_images - val_total_images
    return train_total_images, val_total_images, test_total_images


def assign_groups(
    images_to_resplit: List[Image],
    augmented_images: List[Image],
    groups_info: List[Tuple[str, int, int]],
) -> List[Image]:
    images = []
    for image in augmented_images:
        image.group = "train"
        images.append(image)
    for group_name, images_start, images_end in groups_info:
        for image in images_to_resplit[images_start:images_end]:
            image.group = group_name
            images.append(image)
    return images


def write_json_dataset(dataset: Dataset, target_path=None) -> Dataset:
    to_return = []
    for image in dataset.images:
        to_return.append(
            {
                **image.__dict__,
                "annotations": [annotation.__dict__ for annotation in image.annotations],
            }
        )
    if target_path is None:
        # generate temporary file
        target_path = Path("dataset.json")

    with open(target_path, "w") as file:
        file.write(
            json.dumps(
                {
                    "images": to_return,
                    "categories": [category.__dict__ for category in dataset.categories],
                    "groups": dataset.groups,
                },
                indent=4,
                ensure_ascii=False,
            )
        )
    print("Dataset written to", target_path)

    return dataset


def read_json_dataset(dataset: Dataset | None = None, target_path=None) -> Dataset:
    if target_path is None:
        # generate temporary file
        target_path = Path("dataset.json")

    with open(target_path, "r") as file:
        data = json.load(file)
        images = []
        for image in data["images"]:
            annotations = []
            for annotation in image["annotations"]:
                annotations.append(Annotation(**annotation))
            images.append(Image(**{**image, "annotations": annotations}))  # type: ignore
        categories = []
        for category in data["categories"]:
            categories.append(Category(**category))
        return Dataset(images=images, categories=categories, groups=data["groups"])

    return dataset


def reset_dataset(dataset: Dataset) -> Dataset:
    return Dataset(
        images=[],
        categories=[],
        groups=[],
    )



import math
import tempfile
import colorsys
from dataclasses import asdict
from typing import List
from uuid import uuid4

# from shapely.geometry import Polygon, MultiPolygon, Point, LineString
import cv2
from PIL import Image as ImagePil
from PIL import ImageDraw
from shapely.geometry import Polygon
from shapely.ops import transform
import numpy as np

from pflows.model import get_image_info
from pflows.polygons import (
    bbox_from_polygon,
    calculate_center_from_bbox,
    calculate_center_from_polygon,
)
from pflows.typedef import Annotation, Dataset, Image


# TODO: refactor this function
# pylint: disable=too-many-locals
def crop_to_annotations(dataset: Dataset, tag=None) -> Dataset:
    """
    Crop images to match the min and max x/y of the annotations.
    """
    temp_images_path = tempfile.mkdtemp()
    new_images: List[Image] = []
    for image in dataset.images:
        pil_image = ImagePil.open(image.path)
        width, height = pil_image.size
        min_x = 1000000
        min_y = 1000000
        max_x = 0
        max_y = 0

        for annotation in image.annotations:
            if annotation.bbox is None or (tag is not None and tag not in annotation.tags):
                continue
            min_x = min(int(min(annotation.bbox[0::2]) * width), min_x)
            min_y = min(int(min(annotation.bbox[1::2]) * height), min_y)
            max_x = max(int(max(annotation.bbox[0::2]) * width), max_x)
            max_y = max(int(max(annotation.bbox[1::2]) * height), max_y)
        new_width = int((max_x - min_x))
        new_height = int((max_y - min_y))
        # crop the image
        # Adjust annotations for the cropped image
        for annotation in image.annotations:
            if annotation.bbox is not None and (tag is None or tag in annotation.tags):
                (bbox_x1, bbox_y1, bbox_x2, bbox_y2) = tuple(
                    (
                        (width * coord - min_x) / new_width
                        if i % 2 == 0
                        else (height * coord - min_y) / new_height
                    )
                    for i, coord in enumerate(annotation.bbox)
                )
                annotation.bbox = (bbox_x1, bbox_y1, bbox_x2, bbox_y2)

            if annotation.center is not None:
                (center_x, center_y) = tuple(
                    (
                        (width * coord - min_x) / new_width
                        if i % 2 == 0
                        else (height * coord - min_y) / new_height
                    )
                    for i, coord in enumerate(annotation.center)
                )
                annotation.center = (center_x, center_y)

            if annotation.segmentation is not None:
                new_segmentation = tuple(
                    (
                        (width * coord - min_x) / new_width
                        if i % 2 == 0
                        else (height * coord - min_y) / new_height
                    )
                    for i, coord in enumerate(annotation.segmentation)
                )
                annotation.segmentation = new_segmentation

        cropped_pil_image = pil_image.crop(
            (
                min_x,
                min_y,
                max_x,
                max_y,
            )
        )
        new_path = f"{temp_images_path}/{image.id}.jpg"
        cropped_pil_image.save(new_path)
        new_image = get_image_info(new_path, image.group, image.intermediate_ids + [image.id])
        new_images.append(
            Image(**{**asdict(image), **asdict(new_image), "annotations": image.annotations})
        )
    return Dataset(images=new_images, categories=dataset.categories, groups=dataset.groups)


def create_mask(annotation: Annotation, width: int, height: int):
    mask = np.zeros((height, width), dtype=np.uint8)
    if annotation.segmentation:
        annotations_absolute = [
            value * width if i % 2 == 0 else value * height
            for i, value in enumerate(annotation.segmentation)
        ]
        points = np.array(annotations_absolute, dtype=np.int32).reshape((-1, 2))
        cv2.fillPoly(mask, [points], 1)
    return mask


def crop_mask(mask, crop_x, crop_y, crop_width, crop_height):
    # We need to count the number of mask elements left behind
    # when we crop the mask
    mask_elements = np.sum(mask[crop_y : crop_y + crop_height, crop_x : crop_x + crop_width])
    original_elements = np.sum(mask)
    difference = original_elements - mask_elements
    return mask[crop_y : crop_y + crop_height, crop_x : crop_x + crop_width], bool(difference > 0)


def generate_polygons_from_mask(mask):
    contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    polygons = [Polygon(cnt.reshape(-1, 2)) for cnt in contours if len(cnt) >= 3]
    return polygons


def adjust_polygon_coords(polygon, crop_width, crop_height):
    def adjust_coords(x, y):
        return x / crop_width, y / crop_height

    return transform(adjust_coords, polygon)


def make_id() -> str:
    return str(uuid4())


def get_segmentation_annotations_from_sliced_image(
    sliced_image: Image, original_image: Image, crop_x: int, crop_y: int
) -> List[Annotation]:
    sliced_annotations = []
    w, h = sliced_image.width, sliced_image.height

    for annotation in original_image.annotations or []:
        mask = create_mask(annotation, original_image.width, original_image.height)
        cropped_mask, truncated = crop_mask(mask, crop_x, crop_y, w, h)
        polygons = generate_polygons_from_mask(cropped_mask)

        for polygon in polygons:
            adjusted_polygon = adjust_polygon_coords(polygon, w, h)

            adjusted_segment = tuple(
                float(coord) for point in adjusted_polygon.exterior.coords for coord in point
            )

            sliced_annotation = Annotation(
                id=make_id(),
                category_id=annotation.category_id,
                original_id=annotation.id,
                center=calculate_center_from_polygon(adjusted_segment),
                bbox=annotation.bbox,
                segmentation=tuple(adjusted_segment),
                task=annotation.task,
                conf=annotation.conf,
                category_name=annotation.category_name,
                tags=annotation.tags,
                truncated=truncated,
            )
            sliced_annotations.append(sliced_annotation)

    return sliced_annotations


def get_bbox_annotations_from_sliced_image(
    sliced_image: Image, original_image: Image, crop_x: int, crop_y: int
) -> List[Annotation]:
    sliced_annotations = []
    w, h = sliced_image.width, sliced_image.height

    for annotation in original_image.annotations:
        if annotation.bbox is not None:
            bbox_x1, bbox_y1, bbox_x2, bbox_y2 = annotation.bbox
            # Convert to absolute coordinates
            bbox_x1 = bbox_x1 * original_image.width
            bbox_y1 = bbox_y1 * original_image.height
            bbox_x2 = bbox_x2 * original_image.width
            bbox_y2 = bbox_y2 * original_image.height

            # Check if the bounding box is completely outside the crop
            if (
                bbox_x2 <= crop_x
                or bbox_x1 >= crop_x + w
                or bbox_y2 <= crop_y
                or bbox_y1 >= crop_y + h
            ):
                continue

            # Calculate the intersection of the bounding box with the sliced image
            intersection_x1 = max(bbox_x1, crop_x)
            intersection_y1 = max(bbox_y1, crop_y)
            intersection_x2 = min(bbox_x2, crop_x + w)
            intersection_y2 = min(bbox_y2, crop_y + h)

            # Calculate the adjusted bounding box coordinates
            adjusted_bbox_x1 = max(0, intersection_x1 - crop_x)
            adjusted_bbox_y1 = max(0, intersection_y1 - crop_y)
            adjusted_bbox_x2 = min(w, intersection_x2 - crop_x)
            adjusted_bbox_y2 = min(h, intersection_y2 - crop_y)

            truncated_bbox = False
            if (
                adjusted_bbox_x1 == 0
                or adjusted_bbox_y1 == 0
                or adjusted_bbox_x2 == w
                or adjusted_bbox_y2 == h
            ):
                truncated_bbox = True

            adjusted_bbox_x1 /= w
            adjusted_bbox_y1 /= h
            adjusted_bbox_x2 /= w
            adjusted_bbox_y2 /= h

            sliced_annotation = Annotation(
                id=make_id(),
                category_id=annotation.category_id,
                center=calculate_center_from_bbox(
                    (adjusted_bbox_x1, adjusted_bbox_y1, adjusted_bbox_x2, adjusted_bbox_y2)
                ),
                bbox=(adjusted_bbox_x1, adjusted_bbox_y1, adjusted_bbox_x2, adjusted_bbox_y2),
                segmentation=annotation.segmentation,
                task=annotation.task,
                conf=annotation.conf,
                category_name=annotation.category_name,
                tags=annotation.tags,
                truncated=truncated_bbox,
                original_id=annotation.id,
            )
            sliced_annotations.append(sliced_annotation)

    return sliced_annotations


# pylint: disable=too-many-arguments
def slice_one_image(
    image: Image,
    slice_height: int = 640,
    slice_width: int = 640,
    min_overlap_height_ratio: float = 0.1,
    min_overlap_width_ratio: float = 0.1,
    keep_original_images: bool = False,
) -> List[Image]:

    temp_folder = tempfile.mkdtemp()

    try:
        task_type = image.annotations[0].task
    except:
        task_type = "detect"
    # Only detect or segment is supported for now
    if not task_type in ["detect", "segment", "obb"]:
        return [image]

    sliced_images = []
    img = ImagePil.open(image.path)

    # Calculate the number of slices in each dimension
    num_slices_height = math.ceil(image.height / (slice_height * (1 - min_overlap_height_ratio)))
    num_slices_width = math.ceil(image.width / (slice_width * (1 - min_overlap_width_ratio)))

    if num_slices_height == 1 and num_slices_width == 1:
        if keep_original_images:
            return [image]
        return []
    if num_slices_height == 1:
        num_slices_height = 2
    if num_slices_width == 1:
        num_slices_width = 2
    # Calculate the overlap in pixels
    overlap_height = (num_slices_height * slice_height - image.height) // (num_slices_height - 1)
    overlap_width = (num_slices_width * slice_width - image.width) // (num_slices_width - 1)

    for i in range(num_slices_height):
        for j in range(num_slices_width):
            # Calculate the coordinates for cropping
            left = j * (slice_width - overlap_width)
            top = i * (slice_height - overlap_height)
            right = min(left + slice_width, image.width)
            bottom = min(top + slice_height, image.height)

            sliced_img = img.crop((left, top, right, bottom))
            sliced_path = f"{temp_folder}/slice_{image.id}_{left}_{top}_{right}_{bottom}.jpg"
            if sliced_img.mode == "RGBA":
                sliced_img = sliced_img.convert("RGB")
            sliced_img.save(sliced_path)
            sliced_image = Image(
                id=f"{image.id}_{len(sliced_images)}",
                path=sliced_path,
                intermediate_ids=image.intermediate_ids + [image.id],
                width=slice_width,
                height=slice_height,
                size_kb=sliced_img.size[0] * sliced_img.size[1] // 1024,
                group=image.group,
                tags=image.tags + ["slice"],
                info={
                    "sliced": {
                        "original_path": image.path,
                        "original_width": image.width,
                        "original_height": image.height,
                        "original_image_id": image.id,
                        "left": left,
                        "top": top,
                        "right": right,
                        "bottom": bottom,
                    }
                },
            )
            if task_type in ["segment", "obb"]:
                annotations = get_segmentation_annotations_from_sliced_image(
                    sliced_image, image, left, top
                )
            else:
                annotations = get_bbox_annotations_from_sliced_image(sliced_image, image, left, top)
            sliced_image.annotations = annotations
            sliced_images.append(sliced_image)
    return sliced_images


# def select_best_images_from_slices(sliced_images: List[Image]) -> List[Image]:
#     # Step 1: Count the number of non-truncated annotations for each image
#     image_scores = []
#     images_by_id = {image.id: image for image in sliced_images}
#     for image in sliced_images:
#         non_truncated_count = sum(1 for annotation in image.annotations if not annotation.truncated)
#         image_scores.append((image.id, non_truncated_count))

#     # Step 2: Sort images by the number of non-truncated annotations in descending order
#     image_scores.sort(key=lambda x: x[1], reverse=True)

#     # Step 3: Select images, ensuring all annotations are covered at least once
#     selected_images = set()
#     covered_annotations = set()

#     # Function to get annotations for an image
#     def get_annotations(image_id):
#         for image in sliced_images:
#             if image.id == image_id:
#                 return image.annotations
#         return []

#     all_annotations = {
#         annotation.original_id for image in sliced_images for annotation in image.annotations
#     }

#     for image_id, _ in image_scores:
#         annotations = get_annotations(image_id)
#         current_coverage = {
#             annotation.original_id for annotation in annotations if not annotation.truncated
#         }
#         if not current_coverage.issubset(covered_annotations):
#             selected_images.add(image_id)
#             covered_annotations.update(current_coverage)
#         if covered_annotations == all_annotations:
#             break

#     # Output the selected images
#     return [images_by_id[image_id] for image_id in selected_images]


def is_truncated(annotation: Annotation, image: Image, border_tolerance: int = 10) -> bool:
    if annotation.bbox:
        x1, y1, x2, y2 = annotation.bbox
        return (
            x1 * image.width < border_tolerance
            or y1 * image.height < border_tolerance
            or x2 * image.width > image.width - border_tolerance
            or y2 * image.height > image.height - border_tolerance
        )
    elif annotation.segmentation:
        points = annotation.segmentation
        for i in range(0, len(points), 2):
            x, y = points[i] * image.width, points[i + 1] * image.height
            if (
                x < border_tolerance
                or y < border_tolerance
                or x > image.width - border_tolerance
                or y > image.height - border_tolerance
            ):
                return True
        return False
    return False


def select_best_images_from_slices(
    sliced_images: List[Image], border_tolerance: int = 10
) -> List[Image]:
    # Step 1: Count the number of non-truncated annotations for each image
    image_scores = []
    images_by_id = {image.id: image for image in sliced_images}
    for image in sliced_images:
        non_truncated_count = sum(
            1
            for annotation in image.annotations
            if not is_truncated(annotation, image, border_tolerance)
        )
        image_scores.append((image.id, non_truncated_count))

    # Step 2: Sort images by the number of non-truncated annotations in descending order
    image_scores.sort(key=lambda x: x[1], reverse=True)

    # Step 3: Select images, ensuring all annotations are covered at least once
    selected_images = set()
    covered_annotations = set()

    # Function to get annotations for an image
    def get_annotations(image_id):
        return images_by_id[image_id].annotations

    all_annotations = {
        annotation.original_id for image in sliced_images for annotation in image.annotations
    }

    for image_id, _ in image_scores:
        annotations = get_annotations(image_id)
        current_coverage = {
            annotation.original_id
            for annotation in annotations
            if not is_truncated(annotation, images_by_id[image_id], border_tolerance)
        }
        if not current_coverage.issubset(covered_annotations):
            selected_images.add(image_id)
            covered_annotations.update(current_coverage)
        if covered_annotations == all_annotations:
            break

    # Output the selected images
    return [images_by_id[image_id] for image_id in selected_images]


# pylint: disable=too-many-statements
def join_slices(sliced_images: List[Image]) -> List[Image]:
    # Step 1: Group slices by original image
    slices_by_original_image = {}
    for sliced_image in sliced_images:
        original_image_id = sliced_image.info["sliced"]["original_image_id"]
        if original_image_id not in slices_by_original_image:
            slices_by_original_image[original_image_id] = []
        slices_by_original_image[original_image_id].append(sliced_image)

    # Step 2: Create a new image for each original image
    joined_images = []
    for original_image_id, slices in slices_by_original_image.items():
        first_slice = slices[0]
        original_image_path = first_slice.info["sliced"]["original_path"]
        original_image_width = first_slice.info["sliced"]["original_width"]
        original_image_height = first_slice.info["sliced"]["original_height"]
        # original_image_annotations = original_image.annotations

        joined_image = get_image_info(
            original_image_path,
            first_slice.group,
            [original_image_id] + [slice.id for slice in slices],
        )
        joined_image.tags = first_slice.tags
        joined_image.annotations = []

        already_added = set()
        # Step 3: We need to adjust the annotations for the joined image
        # using the new width and height
        for slice_image in slices:
            for annotation in slice_image.annotations:
                if annotation.id in already_added:
                    continue
                if annotation.truncated:
                    continue
                if annotation.segmentation:
                    segmentation = list(annotation.segmentation)
                    for i in range(0, len(segmentation), 2):
                        x = segmentation[i] * slice_image.width
                        y = segmentation[i + 1] * slice_image.height

                        x += slice_image.info["sliced"]["left"]
                        y += slice_image.info["sliced"]["top"]

                        x /= original_image_width
                        y /= original_image_height

                        segmentation[i] = x
                        segmentation[i + 1] = y

                    tuple_segmentation = tuple(segmentation)
                    annotation.segmentation = tuple_segmentation
                    annotation.center = calculate_center_from_polygon(tuple_segmentation)
                    annotation.bbox = bbox_from_polygon(tuple_segmentation)
                    already_added.add(annotation.id)
                    joined_image.annotations.append(annotation)
                    continue
                if annotation.bbox:
                    x1, y1, x2, y2 = annotation.bbox
                    x1 = x1 * slice_image.width
                    y1 = y1 * slice_image.height
                    x2 = x2 * slice_image.width
                    y2 = y2 * slice_image.height

                    x1 += slice_image.info["sliced"]["left"]
                    y1 += slice_image.info["sliced"]["top"]
                    x2 += slice_image.info["sliced"]["left"]
                    y2 += slice_image.info["sliced"]["top"]

                    x1 /= original_image_width
                    y1 /= original_image_height
                    x2 /= original_image_width
                    y2 /= original_image_height

                    annotation.bbox = (x1, y1, x2, y2)
                    annotation.center = calculate_center_from_bbox(annotation.bbox)
                    joined_image.annotations.append(annotation)
                    already_added.add(annotation.id)
                    continue

        joined_images.append(joined_image)
    return joined_images


def slice_dataset(
    dataset: Dataset,
    slice_height: int = 640,
    slice_width: int = 640,
    overlap_height_ratio: float = 0.1,
    overlap_width_ratio: float = 0.1,
    keep_original_images: bool = False,
    select_best: bool = True,
    border_tolerance: int = 10,
) -> Dataset:
    sliced_images = []
    for image in dataset.images:
        sliced_images.extend(
            slice_one_image(
                image,
                slice_height,
                slice_width,
                overlap_height_ratio,
                overlap_width_ratio,
                keep_original_images,
            )
        )
    if select_best:
        sliced_images = select_best_images_from_slices(sliced_images, border_tolerance)
    return Dataset(images=sliced_images, categories=dataset.categories, groups=dataset.groups)


def join_slice_dataset(
    dataset: Dataset,
) -> Dataset:
    joined_images = join_slices(dataset.images)
    return Dataset(images=joined_images, categories=dataset.categories, groups=dataset.groups)


def write_image_with_annotations(image: Image, target_image_path: str | None = None):
    img = ImagePil.open(image.path)
    img = img.convert("RGB")
    draw = ImageDraw.Draw(img)

    # Generate a color map for different class_ids
    unique_class_ids = set(ann.category_id for ann in image.annotations)
    color_map = {
        class_id: tuple(
            int(x * 255) for x in colorsys.hsv_to_rgb(i / len(unique_class_ids), 1.0, 1.0)
        )
        for i, class_id in enumerate(unique_class_ids)
    }

    for annotation in image.annotations:
        color = color_map[annotation.category_id]
        if annotation.segmentation:
            segmentation = annotation.segmentation
            points = [
                (segmentation[i] * image.width, segmentation[i + 1] * image.height)
                for i in range(0, len(segmentation), 2)
            ]
            draw.polygon(points, outline=color)
        elif annotation.bbox:
            x1, y1, x2, y2 = annotation.bbox
            x1 = x1 * image.width
            y1 = y1 * image.height
            x2 = x2 * image.width
            y2 = y2 * image.height
            draw.rectangle([x1, y1, x2, y2], outline=color, width=2)

    if target_image_path:
        img.save(target_image_path)
    return img


def remove_all_annotations(dataset: Dataset) -> Dataset:
    return Dataset(
        images=[Image(**{**asdict(image), "annotations": []}) for image in dataset.images],
        categories=dataset.categories,
        groups=dataset.groups,
    )
